#
# Unit Tests for Elements of a Block
#

import os
import sys
from os.path import abspath, dirname, join

currdir = dirname( abspath(__file__) )

import pyutilib.th as unittest
import pyutilib.services
from coopr.pyomo import *
from coopr.opt import *
import coopr.environ

solver = load_solvers('glpk')


class TestGenerators(unittest.TestCase):
    
    @ unittest.nottest
    def generate_model(self):
        #
        # ** DO NOT modify the model below without updating the
        # subcomponent_lists and subcomponent_data_lists with the correct
        # declaration orders
        #

        model = ConcreteModel()
        model.q = Set(initialize=[1,2])
        model.Q = Set(model.q,initialize=[1,2])
        model.x = Var(initialize=-1)
        model.X = Var(model.q,initialize=-1)
        model.e = Expression(initialize=-1)
        model.E = Expression(model.q,initialize=-1)
        model.p = Param(mutable=True,initialize=-1)
        model.P = Param(model.q,mutable=True,initialize=-1)
        model.o = Objective(expr=-1)
        model.O = Objective(model.q, rule=lambda model,i: -1)
        model.c = Constraint(expr=model.x>=-1)
        model.C = Constraint(model.q, rule=lambda model,i: model.X[i]>=-1)
        model.sos = SOSConstraint(var=model.X, set=model.q, sos=1)
        model.SOS = SOSConstraint(model.q, var=model.X, set=model.Q, sos=1)
        model.s = Suffix()

        
        model.b = Block()
        model.b.q = Set(initialize=[1,2])
        model.b.Q = Set(model.b.q,initialize=[1,2])
        model.b.x = Var(initialize=0)
        model.b.X = Var(model.b.q,initialize=0)
        model.b.e = Expression(initialize=0)
        model.b.E = Expression(model.b.q,initialize=0)
        model.b.p = Param(mutable=True,initialize=0)
        model.b.P = Param(model.b.q,mutable=True,initialize=0)
        model.b.o = Objective(expr=0)
        model.b.O = Objective(model.b.q, rule=lambda b,i: 0)
        model.b.c = Constraint(expr=model.b.x>=0)
        model.b.C = Constraint(model.b.q, rule=lambda b,i: b.X[i]>=0)
        model.b.sos = SOSConstraint(var=model.b.X, set=model.b.q, sos=1)
        model.b.SOS = SOSConstraint(model.b.q, var=model.b.X, set=model.b.Q, sos=1)
        model.b.s = Suffix()
        model.b.subcomponent_lists = {}
        model.b.subcomponent_data_lists = {}
        model.b.subcomponent_lists[Set] = [model.b.q, model.b.Q]
        model.b.subcomponent_data_lists[Set] = [model.b.q, model.b.Q[1], model.b.Q[2]]
        model.b.subcomponent_lists[Var] = [model.b.x, model.b.X]
        model.b.subcomponent_data_lists[Var] = [model.b.x, model.b.X[1], model.b.X[2]]
        model.b.subcomponent_lists[Expression] = [model.b.e, model.b.E]
        model.b.subcomponent_data_lists[Expression] = [model.b.e, model.b.E[1], model.b.E[2]]
        model.b.subcomponent_lists[Param] = [model.b.p, model.b.P]
        model.b.subcomponent_data_lists[Param] = [model.b.p, model.b.P[1], model.b.P[2]]
        model.b.subcomponent_lists[Objective] = [model.b.o, model.b.O]
        model.b.subcomponent_data_lists[Objective] = [model.b.o, model.b.O[1], model.b.O[2]]
        model.b.subcomponent_lists[Constraint] = [model.b.c, model.b.C]
        model.b.subcomponent_data_lists[Constraint] = [model.b.c, model.b.C[1], model.b.C[2]]
        #model.b.subcomponent_lists[SOSConstraint] = [model.b.sos, model.b.SOS]
        #model.b.subcomponent_data_lists[SOSConstraint] = [model.b.sos, model.b.SOS[1], model.b.SOS[2]]
        model.b.subcomponent_lists[Suffix] = [model.b.s]
        model.b.subcomponent_data_lists[Suffix] = [model.b.s]
        model.b.subcomponent_lists[Block] = []
        model.b.subcomponent_data_lists[Block] = []

        def B_rule(block,i):
            block.q = Set(initialize=[1,2])
            block.Q = Set(block.q,initialize=[1,2])
            block.x = Var(initialize=i)
            block.X = Var(block.q,initialize=i)
            block.e = Expression(initialize=i)
            block.E = Expression(block.q,initialize=i)
            block.p = Param(mutable=True,initialize=i)
            block.P = Param(block.q,mutable=True,initialize=i)
            block.o = Objective(expr=i)
            block.O = Objective(block.q, rule=lambda b,i: i)
            block.c = Constraint(expr=block.x>=i)
            block.C = Constraint(block.q, rule=lambda b,i: b.X[i]>=i)
            block.sos = SOSConstraint(var=block.X, set=block.q, sos=1)
            block.SOS = SOSConstraint(block.q, var=block.X, set=block.Q, sos=1)
            block.s = Suffix()
            block.subcomponent_lists = {}
            block.subcomponent_data_lists = {}
            block.subcomponent_lists[Set] = [block.q, block.Q]
            block.subcomponent_data_lists[Set] = [block.q, block.Q[1], block.Q[2]]
            block.subcomponent_lists[Var] = [block.x, block.X]
            block.subcomponent_data_lists[Var] = [block.x, block.X[1], block.X[2]]
            block.subcomponent_lists[Expression] = [block.e, block.E]
            block.subcomponent_data_lists[Expression] = [block.e, block.E[1], block.E[2]]
            block.subcomponent_lists[Param] = [block.p, block.P]
            block.subcomponent_data_lists[Param] = [block.p, block.P[1], block.P[2]]
            block.subcomponent_lists[Objective] = [block.o, block.O]
            block.subcomponent_data_lists[Objective] = [block.o, block.O[1], block.O[2]]
            block.subcomponent_lists[Constraint] = [block.c, block.C]
            block.subcomponent_data_lists[Constraint] = [block.c, block.C[1], block.C[2]]
            #block.subcomponent_lists[SOSConstraint] = [block.sos, block.SOS]
            #block.subcomponent_data_lists[SOSConstraint] = [block.sos, block.SOS[1], block.SOS[2]]
            block.subcomponent_lists[Suffix] = [block.s]
            block.subcomponent_data_lists[Suffix] = [block.s]
            block.subcomponent_lists[Block] = []
            block.subcomponent_data_lists[Block] = []
        model.B = Block(model.q,rule=B_rule)
        
        model.subcomponent_lists = {}
        model.subcomponent_data_lists = {}
        model.subcomponent_lists[Set] = [model.q, model.Q]
        model.subcomponent_data_lists[Set] = [model.q, model.Q[1], model.Q[2]]
        model.subcomponent_lists[Var] = [model.x, model.X]
        model.subcomponent_data_lists[Var] = [model.x, model.X[1], model.X[2]]
        model.subcomponent_lists[Expression] = [model.e, model.E]
        model.subcomponent_data_lists[Expression] = [model.e, model.E[1], model.E[2]]
        model.subcomponent_lists[Param] = [model.p, model.P]
        model.subcomponent_data_lists[Param] = [model.p, model.P[1], model.P[2]]
        model.subcomponent_lists[Objective] = [model.o, model.O]
        model.subcomponent_data_lists[Objective] = [model.o, model.O[1], model.O[2]]
        model.subcomponent_lists[Constraint] = [model.c, model.C]
        model.subcomponent_data_lists[Constraint] = [model.c, model.C[1], model.C[2]]
        #model.subcomponent_lists[SOSConstraint] = [model.sos, model.SOS]
        #model.subcomponent_data_lists[SOSConstraint] = [model.sos, model.SOS[1], model.SOS[2]]
        model.subcomponent_lists[Suffix] = [model.s]
        model.subcomponent_data_lists[Suffix] = [model.s]
        model.subcomponent_lists[Block] = [model.b, model.B]
        model.subcomponent_data_lists[Block] = [model.b, model.B[1], model.B[2]]
        
        return model

    @ unittest.nottest
    def generator_test(self, ctype):

        model = self.generate_model()

        for block in model.all_blocks(sort_by_keys=True):

            # Non-nested active_subcomponents_generator
            generator = None
            try:
                generator = list(active_subcomponents_generator(block, ctype))
            except:
                if issubclass(ctype, Component):
                    self.fail("active_subcomponents_generator failed with ctype %s" % ctype)
            else:
                if not issubclass(ctype, Component):
                    self.fail("subcomponents_generator should have failed with ctype %s" % ctype)
                self.assertEqual([id(comp) for comp in generator],
                                 [id(comp) for comp in block.subcomponent_lists[ctype]])
            
            # Non-nested subcomponents_generator
            generator = None
            try:
                generator = list(subcomponents_generator(block, ctype))
            except:
                if issubclass(ctype, Component):
                    self.fail("subcomponents_generator failed with ctype %s" % ctype)
            else:
                if not issubclass(ctype, Component):
                    self.fail("subcomponents_generator should have failed with ctype %s" % ctype)
                self.assertEqual([id(comp) for comp in generator],
                                 [id(comp) for comp in block.subcomponent_lists[ctype]])

            # Non-nested active_subcomponents_data_generator, sort_by_keys=True
            generator = None
            try:
                generator = list(active_subcomponents_data_generator(block, ctype, sort_by_keys=True))
            except:
                if issubclass(ctype, Component):
                    self.fail("active_subcomponents_data_generator(sort_by_keys=True) failed with ctype %s" % ctype)
            else:
                if not issubclass(ctype, Component):
                    self.fail("active_subcomponents_data_generator(sort_by_keys=True) should have failed with ctype %s" % ctype)
                self.assertEqual([id(comp) for comp in generator],
                                 [id(comp) for comp in block.subcomponent_data_lists[ctype]])

            # Non-nested active_subcomponents_data_generator, sort_by_keys=False
            generator = None
            try:
                generator = list(active_subcomponents_data_generator(block, ctype, sort_by_keys=False))
            except:
                if issubclass(ctype, Component):
                    self.fail("active_subcomponents_data_generator(sort_by_keys=False) failed with ctype %s" % ctype)
            else:
                if not issubclass(ctype, Component):
                    self.fail("active_subcomponents_data_generator(sort_by_keys=False) should have failed with ctype %s" % ctype)
                self.assertEqual(sorted([id(comp) for comp in generator]),
                                 sorted([id(comp) for comp in block.subcomponent_data_lists[ctype]]))

            # Non-nested sucomponents_data_generator, sort_by_keys=True
            generator = None
            try:
                generator = list(subcomponents_data_generator(block, ctype, sort_by_keys=True))
            except:
                if issubclass(ctype, Component):
                    self.fail("subcomponents_data_generator(sort_by_keys=True) failed with ctype %s" % ctype)
            else:
                if not issubclass(ctype, Component):
                    self.fail("subcomponents_data_generator(sort_by_keys=True) should have failed with ctype %s" % ctype)
                    self.assertEqual([id(comp) for comp in generator],
                                     [id(comp) for comp in block.subcomponent_data_lists[ctype]])

            # Non-nested subcomponents_data_generator, sort_by_keys=False
            generator = None
            try:
                generator = list(subcomponents_data_generator(block, ctype, sort_by_keys=False))
            except:
                if issubclass(ctype, Component):
                    self.fail("subcomponents_data_generator(sort_by_keys=False) failed with ctype %s" % ctype)
            else:
                if not issubclass(ctype, Component):
                    self.fail("subcomponents_data_generator(sort_by_keys=False) should have failed with ctype %s" % ctype)
                self.assertEqual(sorted([id(comp) for comp in generator]),
                                 sorted([id(comp) for comp in block.subcomponent_data_lists[ctype]]))

    def test_Objective(self):
        self.generator_test(Objective)

    def test_Expression(self):
        self.generator_test(Expression)

    def test_Suffix(self):
        self.generator_test(Suffix)

    def test_Constraint(self):
        self.generator_test(Constraint)

    def test_Param(self):
        self.generator_test(Param)

    def test_Var(self):
        self.generator_test(Var)

    #def test_Set(self):
    #    self.generator_test(Set)

    #def test_SOSConstraint(self):
    #    self.generator_test(Set)

    def test_Block(self):

        self.generator_test(Block)

        model = self.generate_model()

        # sorted all_blocks
        self.assertEqual([id(comp) for comp in model.all_blocks(sort_by_keys=True)],
                         [id(comp) for comp in [model,]+model.subcomponent_data_lists[Block]])

        # unsorted all_blocks
        self.assertEqual(sorted([id(comp) for comp in model.all_blocks(sort_by_keys=False)]),
                         sorted([id(comp) for comp in [model,]+model.subcomponent_data_lists[Block]]))

        
class Test(unittest.TestCase):

    def setUp(self):
        #
        # Create block
        #
        self.block = Block()
        self.block.construct()

    def tearDown(self):
        if os.path.exists("unknown.lp"):
            os.unlink("unknown.lp")
        pyutilib.services.TempfileManager.clear_tempfiles()

    def test_clear_attribute(self):
        """ Coverage of the _clear_attribute method """
        obj = Set()
        self.block.A = obj
        self.assertEqual(self.block.A.name,"A")
        self.assertEqual(obj.name,"A")
        self.assertIs(obj, self.block.A)

        obj = Var()
        self.block.A = obj
        self.assertEqual(self.block.A.name,"A")
        self.assertEqual(obj.name,"A")
        self.assertIs(obj, self.block.A)

        obj = Param()
        self.block.A = obj
        self.assertEqual(self.block.A.name,"A")
        self.assertEqual(obj.name,"A")
        self.assertIs(obj, self.block.A)

        obj = Objective()
        self.block.A = obj
        self.assertEqual(self.block.A.name,"A")
        self.assertEqual(obj.name,"A")
        self.assertIs(obj, self.block.A)

        obj = Constraint()
        self.block.A = obj
        self.assertEqual(self.block.A.name,"A")
        self.assertEqual(obj.name,"A")
        self.assertIs(obj, self.block.A)

        obj = Set()
        self.block.A = obj
        self.assertEqual(self.block.A.name,"A")
        self.assertEqual(obj.name,"A")
        self.assertIs(obj, self.block.A)

    def test_set_attr(self):
        self.block.x = Param(mutable=True)
        self.block.x = 5
        self.assertEqual(value(self.block.x), 5)
        self.block.x = 6
        self.assertEqual(value(self.block.x), 6)
        try:
            self.block.x = None
            self.fail("Expected exception assigning None to domain Any")
        except ValueError:
            pass

    def test_display(self):
        self.block.A = RangeSet(1,4)
        self.block.x = Var(self.block.A, bounds=(-1,1))
        def obj_rule(block):
            return summation(block.x)
        self.block.obj = Objective(rule=obj_rule)
        #self.instance = self.block.create()
        #self.block.pprint()
        #self.block.display()

    def Xtest_write2(self):
        self.model.A = RangeSet(1,4)
        self.model.x = Var(self.model.A, bounds=(-1,1))
        def obj_rule(model):
            return summation(model.x)
        self.model.obj = Objective(rule=obj_rule)
        def c_rule(model):
            return (1, model.x[1]+model.x[2], 2)
        self.model.c = Constraint(rule=c_rule)
        self.instance = self.model.create()
        self.instance.write()

    def Xtest_write3(self):
        """Test that the summation works correctly, even though param 'w' has a default value"""
        self.model.J = RangeSet(1,4)
        self.model.w=Param(self.model.J, default=4)
        self.model.x=Var(self.model.J)
        def obj_rule(instance):
            return summation(instance.x, instance.w)
        self.model.obj = Objective(rule=obj_rule)
        self.instance = self.model.create()
        self.assertEqual(len(self.instance.obj[None].expr._args), 4)

    def Xtest_solve1(self):
        self.model.A = RangeSet(1,4)
        self.model.x = Var(self.model.A, bounds=(-1,1))
        def obj_rule(model):
            return summation(model.x)
        self.model.obj = Objective(rule=obj_rule)
        def c_rule(model):
            expr = 0
            for i in model.A:
                expr += i*model.x[i]
            return expr == 0
        self.model.c = Constraint(rule=c_rule)
        self.instance = self.model.create()
        #self.instance.pprint()
        opt = solver['glpk']
        solutions = opt.solve(self.instance, keepfiles=True)
        self.instance.load(solutions)
        self.instance.display(join(currdir,"solve1.out"))
        self.assertFileEqualsBaseline(join(currdir,"solve1.out"),join(currdir,"solve1.txt"))
        #
        def d_rule(model):
            return model.x[1] > 0
        self.model.d = Constraint(rule=d_rule)
        self.model.d.deactivate()
        self.instance = self.model.create()
        solutions = opt.solve(self.instance, keepfiles=True)
        self.instance.load(solutions)
        self.instance.display(currdir+"solve1.out")
        self.assertFileEqualsBaseline(currdir+"solve1.out",currdir+"solve1.txt")
        #
        self.model.d.activate()
        self.instance = self.model.create()
        solutions = opt.solve(self.instance, keepfiles=True)
        self.instance.load(solutions)
        self.instance.display(join(currdir,"solve1.out"))
        self.assertFileEqualsBaseline(join(currdir,"solve1.out"),join(currdir,"solve1a.txt"))
        #
        self.model.d.deactivate()
        def e_rule(i, model):
            return model.x[i] > 0
        self.model.e = Constraint(self.model.A, rule=e_rule)
        self.instance = self.model.create()
        for i in self.instance.A:
            self.instance.e[i].deactivate()
        solutions = opt.solve(self.instance, keepfiles=True)
        self.instance.load(solutions)
        self.instance.display(join(currdir,"solve1.out"))
        self.assertFileEqualsBaseline(join(currdir,"solve1.out"),join(currdir,"solve1b.txt"))

    def Xtest_load1(self):
        """Testing loading of vector solutions"""
        self.model.A = RangeSet(1,4)
        self.model.x = Var(self.model.A, bounds=(-1,1))
        def obj_rule(model):
            return summation(model.x)
        self.model.obj = Objective(rule=obj_rule)
        def c_rule(model):
            expr = 0
            for i in model.A:
                expr += i*model.x[i]
            return expr == 0
        self.model.c = Constraint(rule=c_rule)
        self.instance = self.model.create()
        ans = [0.75]*4
        self.instance.load(ans)
        self.instance.display(join(currdir,"solve1.out"))
        self.assertFileEqualsBaseline(join(currdir,"solve1.out"),join(currdir,"solve1c.txt"))

    def Xtest_solve3(self):
        self.model.A = RangeSet(1,4)
        self.model.x = Var(self.model.A, bounds=(-1,1))
        def obj_rule(model):
            expr = 0
            for i in model.A:
                expr += model.x[i]
            return expr
        self.model.obj = Objective(rule=obj_rule)
        self.instance = self.model.create()
        self.instance.display(join(currdir,"solve1.out"))
        self.assertFileEqualsBaseline(join(currdir,"solve1.out"),join(currdir,"solve3.txt"))

    def Xtest_solve4(self):
        self.model.A = RangeSet(1,4)
        self.model.x = Var(self.model.A, bounds=(-1,1))
        def obj_rule(model):
            return summation(model.x)
        self.model.obj = Objective(rule=obj_rule)
        def c_rule(model):
            expr = 0
            for i in model.A:
                expr += i*model.x[i]
            return expr == 0
        self.model.c = Constraint(rule=c_rule)
        self.instance = self.model.create()
        #self.instance.pprint()
        opt = solver['glpk']
        solutions = opt.solve(self.instance)
        self.instance.load(solutions.solution(0))
        self.instance.display(join(currdir,"solve1.out"))
        self.assertFileEqualsBaseline(join(currdir,"solve1.out"),join(currdir,"solve1.txt"))

if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['display']

import logging
import sys

from six import itervalues

logger = logging.getLogger('coopr.pyomo')


def display(obj, ostream=None):
    """ Display data in a Pyomo object"""
    if ostream is None:
        ostream = sys.stdout
    try:
        display_fcn = obj.display
    except AttributeError:
        raise TypeError(
            "Error trying to display values for object of type %s:\n"
            "\tObject does not support the 'display()' method"
            % (type(obj), ) )
    try:
        display_fcn(ostream=ostream)
    except Exception:
        err = sys.exc_info()[1]
        logger.error(
            "Error trying to display values for object of type %s:\n\t%s"
            % (type(obj), err) )
        raise 


def create_name(name, ndx):
    if ndx is None:
        return name
    if type(ndx) is tuple:
        tmp = str(ndx).replace(', ',',')
        return name+"["+tmp[1:-1]+"]"
    return name+"["+str(ndx)+"]"


def apply_indexed_rule(obj, rule, model, index, options=None):
    try:
        if options is None:
            if index.__class__ is tuple:
                return rule(model, *index)
            elif index is None:
                return rule(model)
            else:
                return rule(model, index)
        else:
            if index.__class__ is tuple:
                return rule(model, *index, **options)
            elif index is None:
                return rule(model, **options)
            else:
                return rule(model, index, **options)
    except TypeError:
        try:
            if options is None:
                return rule(model)
            else:
                return rule(model, **options)
        except:
            # Nothing appears to have matched... re-trigger the original
            # TypeError
            if options is None:
                if index.__class__ is tuple:
                    return rule(model, *index)
                elif index is None:
                    return rule(model)
                else:
                    return rule(model, index)
            else:
                if index.__class__ is tuple:
                    return rule(model, *index, **options)
                elif index is None:
                    return rule(model, **options)
                else:
                    return rule(model, index, **options)

def apply_parameterized_indexed_rule(obj, rule, model, param, index):
    if index.__class__ is tuple:
        return rule(model, param, *index)
    if index is None:
        return rule(model, param)
    return rule(model, param, index)

def _safe_to_str(obj):
    try:
        return str(obj)
    except:
        return "None"

def tabular_writer(ostream, prefix, data, header, row_generator):
    _rows = {}
    #_header = ("Key","Initial Value","Lower Bound","Upper Bound",
    #           "Current Value","Fixed","Stale")
    # NB: _width is a list because we will change these values
    if header:
        _width = [len(x) for x in header]
    else:
        _width = None

    for _key, _val in data:
        try:
            _rows[_key] = tuple( _safe_to_str(x) 
                                 for x in row_generator(_key, _val) )
        except:
            _rows[_key] = None
            continue

        if not _width:
            _width = [0]*len(_rows[_key])
        for _id, x in enumerate(_rows[_key]):
            _width[_id] = max(_width[_id], len(x))

    # NB: left-justify header
    if header:
        # Note: do not right-pad the last header with unnecessary spaces
        tmp = _width[-1]
        _width[-1] = 0
        ostream.write(prefix
                      + " : ".join( "%%-%ds" % _width[i] % x 
                                    for i,x in enumerate(header) )
                      + "\n")
        _width[-1] = tmp

    # If there is no data, we are done...
    if not _rows:
        return

    # right-justify data, except for the last column if there are spaces
    # in the data (probably an expression or vector)
    _width = ["%"+str(i)+"s" for i in _width]

    if sum(1 for x in itervalues(_rows) if x is not None and ' ' in x[-1]):
        _width[-1] = '%s'
    for _key in sorted(_rows):
        _data = _rows[_key]
        if not _data:
            _data = [_key] + [None]*(len(_width)-1)
        ostream.write(prefix
                      + " : ".join( _width[i] % x for i,x in enumerate(_data) )
                      + "\n")




# A simple example illustrating a piecewise
# representation of the step function Z(X)
# 
#          / 0    , 0 <= x <= 1
#  Z(X) >= | 2    , 1 <= x <= 2
#          \ 0.5  , 2 <= x <= 3
#

# **NOTE**: At the domain points 1.0 and 2.0 the
#           range variable can solve to any value
#           on the vertical line. There is no
#           discontinuous "jump".
DOMAIN_PTS = [0., 1., 1., 2., 2., 3.]
RANGE_PTS  = [0., 0., 2., 2., 0.5, 0.5]

from coopr.pyomo import *

model = ConcreteModel()

model.X = Var(bounds=(0,3))
model.Z = Var()

# See documentation on Piecewise component by typing
# help(Piecewise) in a python terminal after importing coopr.pyomo
model.con = Piecewise(model.Z,model.X, # range and domain variables
                      pw_pts=DOMAIN_PTS ,
                      pw_constr_type='EQ',
                      f_rule=RANGE_PTS,
                      pw_repn='INC') # **NOTE**: The not all piecewise represenations
                                     #           handle step functions. Those which do
                                     #           not work with step functions are:
                                     #           BIGM_SOS1, BIGM_BIN, and MC

model.obj = Objective(expr=model.Z+model.X, sense=maximize)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Param']

import sys
import types
import logging
from weakref import ref as weakref_ref
from six import iteritems, iterkeys, next, itervalues

from pyutilib.misc import format_io

from coopr.pyomo.base.component import ComponentData, register_component
from coopr.pyomo.base.sparse_indexed_component import SparseIndexedComponent, normalize_index, UnindexedComponent_set
from coopr.pyomo.base.misc import apply_indexed_rule, apply_parameterized_indexed_rule
from coopr.pyomo.base.numvalue import NumericValue, native_types, value
from coopr.pyomo.base.set_types import Any

logger = logging.getLogger('coopr.pyomo')


class _ParamData(ComponentData, NumericValue):
    """
    This class defines the data for a mutable parameter.

    Constructor Arguments:
        owner       The Param object that owns this data.
        value       The value of this parameter.

    Public Class Attributes:
        value       The numeric value of this variable.
    """

    __slots__ = ('value',)

    def __init__(self, owner, value):
        #
        # The following is equivalent to calling
        # the base ComponentData constructor.
        #
        self._component = weakref_ref(owner)
        #
        # The following is equivalent to calling the
        # base NumericValue constructor.
        #
        self.value = value

    def __getstate__(self):
        """
        This method must be defined because this class uses slots.
        """
        state = super(_ParamData, self).__getstate__()
        for i in _ParamData.__slots__:
            state[i] = getattr(self, i)
        return state

    # Note: because NONE of the slots on this class need to be edited,
    # we don't need to implement a specialized __setstate__ method.

    def clear(self):
        """Clear the data in this component"""
        self.value = None

    def __call__(self, exception=True):
        """
        Return the value of this object.
        """
        return self.value

    def is_fixed(self):
        """
        Returns True because this value is fixed.
        """
        return True

    def is_constant(self):
        """
        Returns False because this is not a constant in an expression.
        """
        return False

    def __nonzero__(self):
        """Return True if the value is defined and non-zero."""
        if self.value:
            return True
        if self.value is None:
            raise ValueError("Param: value is undefined")
        return False

    __bool__ = __nonzero__

    def reset(self):
        """
        Reset the parameter values.  This action does not have
        an effect.
        """
        pass


class Param(SparseIndexedComponent):
    """
    A parameter value, which may be defined over an index.

    Constructor Arguments:
       name        The name of this parameter
       index       The index set that defines the distinct parameters.
                     By default, this is None, indicating that there
                     is a single parameter.
       domain      A set that defines the type of values that
                     each parameter must be.
       within      A set that defines the type of values that
                     each parameter must be.
       validate    A rule for validating this parameter w.r.t. data
                     that exists in the model
       default     A scalar, rule, or dictionary that defines default 
                     values for this parameter
       initialize  A dictionary or rule for setting up this parameter 
                     with existing model data
    """

    DefaultMutable = False

    def __new__(cls, *args, **kwds):
        if cls != Param:
            return super(Param, cls).__new__(cls)
        if args == ():
            return SimpleParam.__new__(SimpleParam)
        else:
            return IndexedParam.__new__(IndexedParam)

    def __init__(self, *args, **kwd):
        self._rule          = kwd.pop('initialize', None )
        self._validate      = kwd.pop('validate', None )
        self.domain         = kwd.pop('domain', Any )
        self.domain         = kwd.pop('within', self.domain )
        self._mutable       = kwd.pop('mutable', Param.DefaultMutable )
        self._default_val   = kwd.pop('default', None )
        #
        if 'repn' in kwd:
            logger.error(
                "The 'repn' keyword is not a validate keyword argument for Param")
        if 'rule' in kwd:
            logger.error(
                "The 'rule' keyword is not a validate keyword argument for Param")
        #
        if self.domain is None:
            self.domain = Any
        #
        kwd.setdefault('ctype', Param)
        SparseIndexedComponent.__init__(self, *args, **kwd)

    def __len__(self):
        """
        Return the number of component data objects stored by this
        component.  If a default value is specified, then the 
        length equals the number of items in the component index.
        """
        if self._default_val is None:
            return len(self._data)
        return len(self._index)

    def __contains__(self, ndx):
        """
        Return true if the index is in the dictionary.  If the default value
        is specified, then all members of the component index are valid.
        """
        if self._default_val is None:
            return ndx in self._data
        return ndx in self._index

    def __iter__(self):
        """
        Iterate over the keys in the dictionary.  If the default value is
        specified, then iterate over all keys in the component index.
        """
        if self._default_val is None:
            return self._data.__iter__()
        return self._index.__iter__()

    #
    # These are "sparse equivalent" access / iteration methods that
    # only loop over the defined data.
    #

    def sparse_keys(self):
        """Return a list of keys in the defined parameters"""
        return list(iterkeys(self._data))

    def sparse_values(self):
        """Return a list of the defined param data objects"""
        return list(itervalues(self._data))

    def sparse_items(self):
        """Return a list (index,data) tuples for defined parameters"""
        return list(iteritems(self._data))

    def sparse_iterkeys(self):
        """Return an iterator for the keys in the defined parameters"""
        return iterkeys(self._data)

    def sparse_itervalues(self):
        """Return an iterator for the defined param data objects"""
        return itervalues(self._data)

    def sparse_iteritems(self):
        """Return an iterator of (index,data) tuples for defined parameters"""
        return iteritems(self._data)

    def extract_values(self):
        """
        A utility to extract all index-value pairs defined for this
        parameter, returned as a dictionary. 

        This method is useful in contexts where key iteration and 
        repeated __getitem__ calls are too expensive to extract 
        the contents of a parameter.
        """
        if self._mutable:
            #
            # The parameter is mutable, parameter data are ParamData types.
            # Thus, we need to create a temporary dictionary that contains the
            # values from the ParamData objects.
            #
            ans = {}
            for key, param_value in self.iteritems():
                ans[key] = param_value.value
            return ans
        elif not self.is_indexed():
            #
            # The parameter is a scalar, so we need to create a temporary 
            # dictionary using the value for this parameter.
            #
            return { None: self.value }
        else:
            #
            # The parameter is not mutable, so iteritems() can be 
            # converted into a dictionary containing parameter values.
            #
            return dict( self.iteritems() )

    def extract_values_sparse(self):
        """
        A utility to extract all index-value pairs defined with non-default
        values, returned as a dictionary. 

        This method is useful in contexts where key iteration and 
        repeated __getitem__ calls are too expensive to extract 
        the contents of a parameter.
        """
        if self._mutable:
            #
            # The parameter is mutable, parameter data are ParamData types.
            # Thus, we need to create a temporary dictionary that contains the
            # values from the ParamData objects.
            #
            ans = {}
            for key, param_value in self.sparse_iteritems():
                ans[key] = param_value.value
            return ans
        elif not self.is_indexed():
            #
            # The parameter is a scalar, so we need to create a temporary 
            # dictionary using the value for this parameter.
            #
            return { None: self.value }
        else:
            #
            # The parameter is not mutable, so sparse_iteritems() can be 
            # converted into a dictionary containing parameter values.
            #
            return dict( self.sparse_iteritems() )

    def store_values(self, new_values, check=True):
        """
        A utility to update a Param with a dictionary or scalar.

        If check=True, then both the index and value  
        are checked through the __getitem__ method.  Using check=False
        should only be used by developers!
        """
        if not self._mutable:
            raise RuntimeError("Cannot call store_values method on immutable Param="+ self.cname(True))
        #
        _srcType = type(new_values)
        _isDict = _srcType is dict or ( \
            hasattr(_srcType, '__getitem__') 
            and not isinstance(new_values, NumericValue) )
        # 
        if check is True:
            if _isDict:
                for index, new_value in iteritems(new_values):
                    self[index] = new_value
            else:
                for index in self._index:
                    self[index] = new_values
            return
        #
        # The argument check is False, so we bypass almost all of the
        # Param logic for ensuring data integrity.
        #
        if self.is_indexed():
            if _isDict:
                # It is possible that the Param is sparse and that the
                # index is not already in the _data dict.  As these
                # cases are rare, we will recover from the exception
                # instead of incurring the penalty of checking.
                for index, new_value in iteritems(new_values):
                    try:
                        self._data[index].value = new_value
                    except:
                        self._data[index] = _ParamData(self, new_value)
            else:
                # For scalars, we will choose an approach based on
                # how "dense" the Param is
                if not self._data: # empty
                    for index in self._index:
                        self._data[index] = _ParamData(self, new_values)
                elif len(self._data) == len(self._index):
                    for index in self._index:
                        self._data[index].value = new_values
                else:
                    for index in self._index:
                        if index in self._data:
                            self._data[index].value = new_values
                        else:
                            self._data[index] = _ParamData(self, new_values)
        else:
            #
            # Initialize a scalar
            #
            if _isDict:
                if None not in new_values:
                    raise RuntimeError(
                        "Cannot store value for scalar Param="+
                        self.cname(True)+"; no value with index None "
                        "in input new values map.")
                new_values = new_values[None]
            # scalars have to be handled differently
            self._data[None] = new_values

    def _default(self, idx):
        """
        Returns the default component data value
        """
        #
        # Local values
        #
        val = self._default_val
        _default_type = type(val)
        #
        if not self._constructed:
            if idx is None:
                idx_str = '%s' % (self.name,)
            else:
                idx_str = '%s[%s]' % (self.name, idx,)
            raise ValueError(
                "Error retrieving Param value (%s): This parameter has "
                "not been constructed" % ( idx_str,) )                
        if val is None:
            if self.is_indexed():
                idx_str = '%s[%s]' % (self.cname(True), idx,)
            else:
                idx_str = '%s' % (self.cname(True),)
            raise ValueError(
                    "Error retrieving Param value (%s): The Param value is "
                    "undefined and no default value is specified"
                    % ( idx_str,) )
        #
        # Get the value of the default for this index
        #
        _check_value_domain = True
        if _default_type in native_types:
            #
            # The set_default() method validates the domain of native types, so
            # we can skip the check on the value domain.
            #
            _check_value_domain = False
        elif _default_type is types.FunctionType:
            val = apply_indexed_rule(self, val, self.parent(), idx)
        elif hasattr(val, '__getitem__') and not isinstance(val, NumericValue):
            val = val[idx]
        else:
            pass
        if _check_value_domain:
            #
            # Get the value of a numeric value
            #
            if val.__class__ not in native_types:
                if isinstance(val, NumericValue):
                    val = val()
            #
            # Check the domain
            # 
            if val not in self.domain:
                raise ValueError(
                    "Invalid default parameter value: %s[%s] = '%s';"
                    " value type=%s.\n\tValue not in parameter domain %s" %
                    (self.cname(True), idx, val, type(val), self.domain.name) )
        #
        # Set the parameter
        #  
        if self._mutable:
            if idx is None:
                self._raw_setitem(idx, val)
            else:
                self._data[idx] = _ParamData(self, val)
                #self._raw_setitem(idx, _ParamData(self, val), True)
            return self[idx]
        else:
            #
            # This is kludgy: If the user wants to validate the Param
            # values, we need to validate the default value as well.
            # For Mutable Params, this is easy: setitem will inject the
            # value into _data and then call validate.  For immutable
            # params, we never inject the default into the data
            # dictionary.  This will break validation, as the validation
            # rule is allowed to assume the data is already present
            # (actually, it will die on infinite recursion, as
            # Param.__getitem__() will re-call _default).
            #
            # So, we will do something very inefficient: if we are
            # validating, we will inject the value into the dictionary,
            # call validate, and remove it.
            #
            if self._validate:
                try:
                    self._data[idx] = val
                    self._validateitem(idx, val)
                finally:
                    del self._data[idx]
            return val

    def set_default(self, val):
        """
        Perform error checks and then set the default value for this parameter.

        NOTE: this test will not validate the value of function return values.
        """
        if self._constructed \
                and val is not None \
                and type(val) in native_types \
                and val not in self.domain:
            raise ValueError(
                "Default value (%s) is not valid for Param domain %s" %
                ( str(val), self.domain.name ) )
        self._default_val = val

    def default(self):
        """
        Return the value of the parameter default.

        Possible values:
            None            No default value is provided.
            Numeric         A constant value that is the default value for all
                                undefined parameters.
            Function        f(model, i) returns the value for the default value
                                for parameter i
        """
        return self._default_val

    def _raw_setitem(self, ndx, val):
        """
	    The __setitem__ method performs significant validation
	    around the input indices, particularly when the index
	    value is new.  In various contexts, we don't need to
	    incur this overhead (e.g. during initialization).  The
	    _raw_setitem assumes the input value is in the set
	    native_types
        """
        #
        # Params should contain *values*.  Note that if we just call 
        # value(), then that forces the value to be a numeric value.
        # Notably, we allow Params with domain==Any to hold strings, tuples,
        # etc.  The following lets us use NumericValues to initialize
        # Params, but is optimized to check for "known" native types to
        # bypass a potentially expensive isinstance()==False call.
        #
        if val.__class__ not in native_types:
            if isinstance(val, NumericValue):
                val = val()
        #
        # Check if the value is valid within the current domain
        #
        if val not in self.domain:
            raise ValueError(
                "Invalid parameter value: %s[%s] = '%s', value type=%s.\n"
                "\tValue not in parameter domain %s" %
                (self.cname(True), ndx, val, type(val), self.domain.name) )
        #
        # Set the value depending on the type of param value.
        #
        if ndx is None:
            self.value = val
        elif self._mutable:
            if ndx in self._data:
                self._data[ndx].value = val
            else:
                self._data[ndx] = _ParamData(self, val)
        else:
            self._data[ndx] = val
        #
        # Execute a validation operation
        #
        if self._validate:
            self._validateitem(ndx, val)

    def _validateitem(self, ndx, val):
        """
        Validate a given input/value pair.
        """
        if not apply_parameterized_indexed_rule(self, self._validate, self.parent(), val, ndx):
            raise ValueError(
                "Invalid parameter value: %s[%s] = '%s', value type=%s.\n"
                "\tValue failed parameter validation rule" %
                ( self.cname(True), ndx, val, type(val) ) )

    def __setitem__(self, ndx, val):
        """
        Add a parameter value to the index.
        """
        #
        # TBD: Potential optimization: if we find that updating a Param is
        # more common than setting it in the first place, then first
        # checking the _data and then falling back on the _index *might*
        # be more efficient.
        #
        if self._constructed and not self._mutable:
            raise TypeError(
"""Attempting to set the value of the immutable parameter %s after the
parameter has been constructed.  If you intend to change the value of
this parameter dynamically, please declare the parameter as mutable
[i.e., Param(mutable=True)]""" % (self.cname(True),))
        #
        # Check if we have a valid index.
        # We assume that most calls to this method will send either a
        # scalar or a valid tuple.  So, for efficiency, we will check the
        # index *first*, and only go through the hassle of
        # flattening things if the ndx is not found.
        #
        if ndx not in self._index:
            ndx = normalize_index(ndx)
            if ndx not in self._index:
                if not self.is_indexed():
                    msg = "Error setting parameter value: " \
                          "Cannot treat the scalar Param '%s' as an array" \
                          % ( self.cname(True), )
                else:
                    msg = "Error setting parameter value: " \
                          "Index '%s' is not valid for array Param '%s'" \
                          % ( ndx, self.cname(True), )
                raise KeyError(msg)
        # We have a valid index, so do the actual set operation.
        self._raw_setitem(ndx, val)

    def _initialize_from(self, _init):
        """
        Initialize data from a rule or data
        """
        _init_type = type(_init)
        _isDict = _init_type is dict

        if _isDict or _init_type in native_types:
            #
            # We skip the other tests if we have a dictionary or constant
            # value, as these are the most common cases.
            #
            pass

        elif _init_type is types.FunctionType:
            #
            # Initializing from a function
            #
            if not self.is_indexed():
                #
                # A scalar value has a single value.
                # We call __setitem__, which does checks on the value.
                #
                self[None] = _init(self.parent())
                return
            else:
                #
                # An indexed parameter, where we call the function for each
                # index.
                #
                self_parent = self.parent()
                #
                try:
                    #
                    # Create an iterator for the indices.  We assume that 
                    # it returns flattened tuples. Otherwise,
                    # the validation process is far too expensive.
                    #
                    _iter = self._index.__iter__()
                    idx = next(_iter)
                    #
                    # If a function returns a dict (or
                    # dict-like thing), then we initialize the Param object
                    # by reseting _init and _isDict
                    #
                    # Note that this logic allows the user to call a
                    # function without an index
                    #
                    val = apply_indexed_rule(self, _init, self_parent, idx)

                    # 
                    # The following is a simplification of the main
                    # _initialize_from logic.  The idea is that if the
                    # function returns a scalar-like thing, use it to
                    # initialize this index and re-call the function for
                    # the next value.  However, if the function returns
                    # somethign that is dict-like, then use the dict to
                    # initialize everything and do not re-vall the
                    # initialize function.
                    #
                    # Note: while scalar components are technically
                    # "dict-like", we will treat them as scalars and
                    # re-call the initialize function.
                    #
                    _dict_like = False
                    if type(val) is dict:
                        _dict_like = True
                    elif isinstance(val, SparseIndexedComponent):
                        _dict_like = val.is_indexed()
                    elif hasattr(val, '__getitem__') \
                            and not isinstance(val, NumericValue):
                        try:
                            for x in _init:
                                _init.__getitem__(x)
                            _dict_like = True
                        except:
                            pass

                    if _dict_like:
                        _init = val
                        _isDict = True
                    else:
                        #
                        # At this point, we know the value is specific to
                        # this index (i.e., not likely to be a dict-like
                        # thing), and that the index is valid; so, it is
                        # safe to use _raw_setitem (which will perform all
                        # the domain / validation checking)
                        #
                        self._raw_setitem(idx, val)
                        #
                        # Now iterate over the rest of the index set.
                        # 
                        for idx in _iter:
                            self._raw_setitem( idx, 
                                apply_indexed_rule( self, _init, self_parent, idx ) )
                        return
                except StopIteration:
                    #
                    # The index set was empty...
                    # The parameter is indexed by an empty set, or an empty set tuple. 
                    # Rare, but it has happened.
                    #
                    return

        elif isinstance(_init, NumericValue):
            #
            # Reduce NumericValues to scalars.  This allows us to treat 
            # scalar components as numbers and not
            # as indexed components with a index set of [None]
            #
            _init = _init()

        elif isinstance(_init, SparseIndexedComponent):
            #
            # Ideally, we want to reduce SparseIndexedComponents to
            # a dict, but without "densifying" it.  However, since
            # there is no way to (easily) get the default value, we
            # will take the "less surprising" route of letting the
            # source become dense, so that we get the expected copy.
            #
            # TODO: Establish use-cases where we can use default values
            # for sparsity.
            #
            sparse_src = len(_init) != len(_init.keys())
            tmp = dict( _init.iteritems() )
            if sparse_src and len(_init) == len(_init.keys()):
                logger.warning("""
Initializing Param %s using a sparse mutable indexed component (%s).
This has resulted in the conversion of the source to dense form.
""" % ( self.cname(True), _init.name ) )
            _init = tmp
            _isDict = True

        #
        # If the _init is not a native dictionary, but it
        # behaves like one (that is, it could be converted to a
        # dict with "dict((key,_init[key]) for key in _init)"),
        # then we will treat it as such
        #
        # TODO: Establish a use-case for this.  This iteration is
        # expensive.
        #
        if not _isDict and hasattr(_init, '__getitem__'):
            try:
                _isDict = True
                for x in _init:
                    _init.__getitem__(x)
            except:
                _isDict = False
        #
        # Now, we either have a scalar or a dictionary
        #
        if _isDict:
            #
            # Because this is a user-specified dictionary, we
            # must use the normal (expensive) __setitem__ route
            # so that the individual indices are validated.
            #
            for key in _init:
                self[key] = _init[key]
        else:
            try:
                #
                # A constant is being supplied as a default to
                # a parameter.  This happens for indexed parameters,
                # particularly when dealing with mutable parameters.
                #
                # We look at the first iteration index separately to
                # to validate the value against the domain once.
                #
                _iter = self._index.__iter__()
                idx = next(_iter)
                self._raw_setitem(idx, _init)
                #
                # Note: the following is safe for both indexed and
                # non-indexed parameters: for non-indexed, the first
                # idx (above) will be None, and the for-loop below
                # will NOT be called.
                #
                if self._mutable:
                    _init = self[idx].value
                    for idx in _iter:
                        self._raw_setitem( idx, _ParamData(self,_init) )
                else:
                    _init = self[idx]
                    for idx in _iter:
                        self._raw_setitem(idx, _init)
            except StopIteration:
                #
                # The index set was empty...
                # The parameter is indexed by an empty set, or an empty set tuple. 
                # Rare, but it has happened.
                #
                pass

    def construct(self, data=None):
        """
        Initialize this component.

	    A parameter is constructed using the initial data or
	    the data loaded from an external source.  We first set
	    all the values based on self._rule, and then allow the
	    data dictionary to overwrite anything.

	    Note that we allow an undefined Param value to be
	    constructed.  We throw an exception if a user tries to
	    use an uninitialized Param.
        """
        if __debug__:   #pragma:nocover
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing Param, name=%s, from data=%s"
                             % ( self.cname(True), str(data) ))
        #
        if self._constructed:
            return
        #
        # If the default value is a simple type, we check it versus
        # the domain.
        #
        val = self._default_val
        if val is not None \
                and type(val) in native_types \
                and val not in self.domain:
            raise ValueError(
                "Default value (%s) is not valid for Param domain %s" %
                ( str(val), self.domain.name ) )
        #
        # Step #1: initialize data from rule value
        #
        if self._rule is not None:
            self._initialize_from(self._rule)
        #
        # Step #2: allow any user-specified (external) data to override
        # the initialization
        #
        if data is not None:
            try:
                for key, val in iteritems(data):
                    self[key] = val
            except Exception:
                msg = sys.exc_info()[1]
                if type(data) is not dict:
                   raise ValueError(
                       "Attempting to initialize parameter=%s with data=%s.\n"
                       "\tData type is not a dictionary, and a dictionary is "
                       "expected." % (self.cname(True), str(data)) )
                else:
                    raise RuntimeError(
                        "Failed to set value for param=%s, index=%s, value=%s."
                        "\n\tsource error message=%s" 
                        % (self.cname(True), str(key), str(val), str(msg)) )
        #
        self._constructed = True

    def reconstruct(self, data=None):
        """
        Reconstruct this parameter object.  This is particularly useful
        for cases where an initialize rule is provided.  An initialize
        rule can return an expression that is a function of other
        parameters, so reconstruction can account for changes in dependent
        parameters.

        Only mutable parameters can be reconstructed.  Otherwise, the
        changes would not be propagated into expressions in objectives
        or constraints.
        """
        if not self._mutable:
            raise RuntimeError("Cannot invoke reconstruct method of immutable param="+self.cname(True))
        SparseIndexedComponent.reconstruct(self, data=data)

    def _pprint(self):
        """
        Return data that will be printed for this component.
        """
        return ( [("Size", len(self)),
                  ("Index", self._index \
                       if self._index != UnindexedComponent_set else None),
                  ("Domain", self.domain.name),
                  ("Default", "(function)" if type(self._default_val) \
                       is types.FunctionType else self._default_val),
                  ("Mutable", self._mutable),
                  ],
                 self.sparse_iteritems(),
                 ("Key","Value"),
                 lambda k, v: [ k,
                                value(v)
                                ]
                 )


class SimpleParam(_ParamData, Param):

    def __init__(self, *args, **kwds):
        Param.__init__(self, *args, **kwds)
        _ParamData.__init__(self, self, kwds.get('default',None))
        self._data[None] = self

    #
    # Since this class derives from Component and Component.__getstate__
    # just packs up the entire __dict__ into the state dict, there s
    # nothng special that we need to do here.  We will just defer to the
    # super() get/set state.  Since all of our get/set state methods
    # rely on super() to traverse the MRO, this will automatically pick
    # up both the Component and Data base classes.
    #

    def __call__(self, exception=True):
        """
        Return the value of this parameter.
        """
        if self._constructed:
            return _ParamData.__call__(self, exception=exception)
        if exception:
            raise ValueError( """Evaluating the numeric value of parameter '%s' before the Param has been
            constructed (there is currently no value to return).""" % self.cname(True) )

    def set_value(self, value):
        if self._constructed and not self._mutable:
            raise TypeError(
"""Attempting to set the value of the immutable parameter %s after the
parameter has been constructed.  If you intend to change the value of
this parameter dynamically, please declare the parameter as mutable
[i.e., Param(mutable=True)]""" % (self.cname(True),))
        self[None] = value
        
    def is_constant(self):
        """
        Returns False because this is not a constant in an expression.
        """
        return self._constructed and not self._mutable


class IndexedParam(Param):

    def __call__(self, exception=True):
        """Compute the value of the parameter"""
        if exception:
            msg = 'Cannot compute the value of an array of parameters'
            raise TypeError(msg)

    def reset(self):
        """
        Reset the parameter values.  This action does not have
        an effect.
        """
        pass                    #pragma:nocover


register_component(Param, "Parameter data that is used to define a model instance.")




#
# Unit Tests for Elements of a Model
#
# TestSimpleObj                Class for testing single objective
# TestArrayObj                Class for testing array of objective
#

import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

from coopr.pyomo.base import IntegerSet
from coopr.pyomo import *
from coopr.opt import *
import pyutilib.th as unittest
import pyutilib.services

class TestSimpleObj(unittest.TestCase):

    def test_numeric_expr(self):
        """Test expr option with a single numeric constant"""
        model = ConcreteModel()
        model.obj = Objective(expr=0.0)
        self.assertEqual(model.obj(), 0.0)
        self.assertEqual(value(model.obj), 0.0)
        self.assertEqual(value(model.obj._data[None]), 0.0)

    def test_mutable_param_expr(self):
        """Test expr option with a single mutable param"""
        model = ConcreteModel()
        model.p = Param(initialize=1.0,mutable=True)
        model.obj = Objective(expr=model.p)
        
        self.assertEqual(model.obj(), 1.0)
        self.assertEqual(value(model.obj), 1.0)
        self.assertEqual(value(model.obj._data[None]), 1.0)

    def test_immutable_param_expr(self):
        """Test expr option a single immutable param"""
        model = ConcreteModel()
        model.p = Param(initialize=1.0,mutable=False)
        model.obj = Objective(expr=model.p)
        
        self.assertEqual(model.obj(), 1.0)
        self.assertEqual(value(model.obj), 1.0)
        self.assertEqual(value(model.obj._data[None]), 1.0)

    def test_var_expr(self):
        """Test expr option with a single var"""
        model = ConcreteModel()
        model.x = Var(initialize=1.0)
        model.obj = Objective(expr=model.x)
        
        self.assertEqual(model.obj(), 1.0)
        self.assertEqual(value(model.obj), 1.0)
        self.assertEqual(value(model.obj._data[None]), 1.0)

    def test_expr1_option(self):
        """Test expr option"""
        model = ConcreteModel()
        model.x = Var(RangeSet(1,4),initialize=2)
        ans=0
        for i in model.x.keys():
            ans = ans + model.x[i]
        model.obj = Objective(expr=ans)
        
        self.assertEqual(model.obj(), 8)
        self.assertEqual(value(model.obj), 8)
        self.assertEqual(value(model.obj._data[None]), 8)

    def test_expr2_option(self):
        """Test expr option"""
        model = ConcreteModel()
        model.x = Var(initialize=2)
        model.obj = Objective(expr=model.x)
        
        model.x.reset()
        #print 'X',type(model.obj.rule)
        self.assertEqual(model.obj(), 2)
        self.assertEqual(value(model.obj), 2)
        self.assertEqual(value(model.obj._data[None]), 2)

    def test_rule_option(self):
        """Test rule option"""
        model = ConcreteModel()
        def f(model):
            ans=0
            for i in model.x.keys():
                ans = ans + model.x[i]
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.obj = Objective(rule=f)
        
        self.assertEqual(model.obj(), 8)
        self.assertEqual(value(model.obj), 8)
        self.assertEqual(value(model.obj._data[None]), 8)

    def test_arguments(self):
        """Test that arguments notare of type SimpleSet"""
        model = ConcreteModel()
        def rule(model):
            return 1
        try:
            model.obj = Objective(model, rule=rule)
        except TypeError:
            pass
        else:
            self.fail("Objective should only accept SimpleSets")

    def test_sense_option(self):
        """Test sense option"""
        model = ConcreteModel()
        def rule(model):
            return 1.0
        model.obj = Objective(sense=maximize, rule=rule)
        
        self.assertEqual(model.obj.sense, maximize)
        self.assertEqual(model.obj.is_minimizing(), False)

    def test_dim(self):
        """Test dim method"""
        model = ConcreteModel()
        def rule(model):
            return 1
        model.obj = Objective(rule=rule)
        
        self.assertEqual(model.obj.dim(),0)

    def test_keys(self):
        """Test keys method"""
        model = ConcreteModel()
        def rule(model):
            return 1
        model.obj = Objective(rule=rule)
        
        self.assertEqual(list(model.obj.keys()),[None])

    def test_len(self):
        """Test len method"""
        model = ConcreteModel()
        def rule(model):
            return 1.0
        model.obj = Objective(rule=rule)
        
        self.assertEqual(len(model.obj),1)
        """Test rule option"""
        def f(model):
            ans=0
            for i in model.x.keys():
                ans = ans + model.x[i]
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.obj = Objective(rule=f)
        
        self.assertEqual(len(model.obj),1)


class TestArrayObj(unittest.TestCase):

    def create_model(self):
        #
        # Create Model
        #
        model = ConcreteModel()
        model.A = Set(initialize=[1,2])
        return model

    def test_rule_option1(self):
        """Test rule option"""
        model = self.create_model()
        def f(model, i):
            ans=0
            for j in model.x.keys():
                ans = ans + model.x[j]
            ans *= i
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.obj = Objective(model.A,rule=f)
        
        self.assertEqual(model.obj[1](), 8)
        self.assertEqual(model.obj[2](), 16)
        self.assertEqual(value(model.obj[1]), 8)
        self.assertEqual(value(model.obj[2]), 16)

    def test_rule_option2(self):
        """Test rule option"""
        model = self.create_model()
        def f(model, i):
            if i == 1:
                return Objective.Skip
            ans=0
            for j in model.x.keys():
                ans = ans + model.x[j]
            ans *= i
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.obj = Objective(model.A,rule=f)
        
        self.assertEqual(model.obj[2](), 16)
        self.assertEqual(value(model.obj[2]), 16)

    def test_rule_option3(self):
        """Test rule option"""
        model = self.create_model()
        @simple_objective_rule
        def f(model, i):
            if i == 1:
                return None
            ans=0
            for j in model.x.keys():
                ans = ans + model.x[j]
            ans *= i
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.obj = Objective(model.A,rule=f)
        
        self.assertEqual(model.obj[2](), 16)
        self.assertEqual(value(model.obj[2]), 16)

    def test_rule_numeric_expr(self):
        """Test rule option with returns a single numeric constant for the expression"""
        model = self.create_model()
        def f(model, i):
            return 1.0
        model.obj = Objective(model.A,rule=f)
        
        self.assertEqual(model.obj[2](), 1.0)
        self.assertEqual(value(model.obj[2]), 1.0)

    def test_rule_immutable_param_expr(self):
        """Test rule option that returns a single immutable param for the expression"""
        model = self.create_model()
        def f(model, i):
            return model.p[i]
        model.p = Param(RangeSet(1,4),initialize=1.0,mutable=False)
        model.x = Var()
        model.obj = Objective(model.A,rule=f)
        
        self.assertEqual(model.obj[2](), 1.0)
        self.assertEqual(value(model.obj[2]), 1.0)

    def test_rule_mutable_param_expr(self):
        """Test rule option that returns a single mutable param for the expression"""
        model = self.create_model()
        def f(model, i):
            return model.p[i]
        model.r = RangeSet(1,4)
        model.p = Param(model.r,initialize=1.0,mutable=True)
        model.x = Var()
        model.obj = Objective(model.A,rule=f)
        
        self.assertEqual(model.obj[2](), 1.0)
        self.assertEqual(value(model.obj[2]), 1.0)

    def test_rule_var_expr(self):
        """Test rule option that returns a single var for the expression"""
        model = self.create_model()
        def f(model, i):
            return model.x[i]
        model.r = RangeSet(1,4)
        model.x = Var(model.r,initialize=1.0)
        model.obj = Objective(model.A,rule=f)
        
        self.assertEqual(model.obj[2](), 1.0)
        self.assertEqual(value(model.obj[2]), 1.0)

    def test_sense_option(self):
        """Test sense option"""
        model = self.create_model()
        model.obj = Objective(model.A,sense=maximize)
        
        self.assertEqual(model.obj.sense, maximize)
        self.assertEqual(model.obj.is_minimizing(), False)
        for i in model.obj:
            self.assertEqual(model.obj[i].is_minimizing(), False)

    def test_dim(self):
        """Test dim method"""
        model = self.create_model()
        model.obj = Objective(model.A)
        
        self.assertEqual(model.obj.dim(),1)

    def test_keys(self):
        """Test keys method"""
        model = self.create_model()
        def A_rule(model, i):
            return model.x
        model.x = Var()
        model.obj = Objective(model.A, rule=A_rule)
        
        self.assertEqual(len(model.obj.keys()),2)

    def test_len(self):
        """Test len method"""
        model = self.create_model()
        model.obj = Objective(model.A)
        
        self.assertEqual(len(model.obj),0)
        """Test rule option"""
        def f(model):
            ans=0
            for i in model.x.keys():
                ans = ans + model.x[i]
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.obj = Objective(rule=f)
        
        self.assertEqual(len(model.obj),1)


class Test2DArrayObj(unittest.TestCase):

    def create_model(self):
        model = ConcreteModel()
        model.A = Set(initialize=[1,2])
        return model

    def test_rule_option1(self):
        """Test rule option"""
        model = self.create_model()
        def f(model, i, k):
            ans=0
            for j in model.x.keys():
                ans = ans + model.x[j]
            ans *= i
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.obj = Objective(model.A,model.A, rule=f)
        
        try:
            self.assertEqual(model.obj(),None)
            self.fail("Expected TypeError")
        except TypeError:
            pass
        model.x.reset()
        self.assertEqual(model.obj[1,1](), 8)
        self.assertEqual(model.obj[2,1](), 16)
        self.assertEqual(value(model.obj[1,1]), 8)
        self.assertEqual(value(model.obj[2,1]), 16)

    def test_sense_option(self):
        """Test sense option"""
        model = self.create_model()
        model.obj = Objective(model.A,model.A,sense=maximize)
        
        self.assertEqual(model.obj.sense, maximize)
        self.assertEqual(model.obj.is_minimizing(), False)
        for i in model.obj:
            self.assertEqual(model.obj[i].is_minimizing(), False)

    def test_dim(self):
        """Test dim method"""
        model = self.create_model()
        model.obj = Objective(model.A,model.A)
        
        self.assertEqual(model.obj.dim(),2)

    def test_keys(self):
        """Test keys method"""
        model = self.create_model()
        def A_rule(model, i, j):
            return model.x
        model.x = Var()
        model.obj = Objective(model.A,model.A, rule=A_rule)
        
        self.assertEqual(len(model.obj.keys()),4)

    def test_len(self):
        """Test len method"""
        model = self.create_model()
        model.obj = Objective(model.A,model.A)
        
        self.assertEqual(len(model.obj),0)
        """Test rule option"""
        def f(model):
            ans=0
            for i in model.x.keys():
                ans = ans + model.x[i]
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.obj = Objective(rule=f)
        
        self.assertEqual(len(model.obj),1)


class TestObjList(unittest.TestCase):

    def create_model(self):
        model = ConcreteModel()
        model.A = Set(initialize=[1,2,3,4])
        return model

    def test_rule_option1(self):
        """Test rule option"""
        model = self.create_model()
        def f(model, i):
            if i > 4:
                return ObjectiveList.End
            ans=0
            for j in model.x:
                ans = ans + model.x[j]
            ans *= i
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.o = ObjectiveList(rule=f)

        self.assertEqual(model.o[1](), 8)
        self.assertEqual(model.o[2](), 16)
        self.assertEqual(len(model.o), 4)

    def test_rule_option2(self):
        """Test rule option"""
        model = self.create_model()
        def f(model, i):
            if i > 2:
                return ObjectiveList.End
            i = 2*i - 1
            ans=0
            for j in model.x:
                ans = ans + model.x[j]
            ans *= i
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.o = ObjectiveList(rule=f)

        self.assertEqual(model.o[1](), 8)
        self.assertEqual(len(model.o), 2)

    def test_rule_option1a(self):
        """Test rule option"""
        model = self.create_model()
        @simple_objectivelist_rule
        def f(model, i):
            if i > 4:
                return None
            ans=0
            for j in model.x:
                ans = ans + model.x[j]
            ans *= i
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.o = ObjectiveList(rule=f)

        self.assertEqual(model.o[1](), 8)
        self.assertEqual(model.o[2](), 16)
        self.assertEqual(len(model.o), 4)

    def test_rule_option2a(self):
        """Test rule option"""
        model = self.create_model()
        @simple_objectivelist_rule
        def f(model, i):
            if i > 2:
                return None
            i = 2*i - 1
            ans=0
            for j in model.x:
                ans = ans + model.x[j]
            ans *= i
            return ans
        model.x = Var(RangeSet(1,4),initialize=2)
        model.o = ObjectiveList(rule=f)

        self.assertEqual(model.o[1](), 8)
        self.assertEqual(len(model.o), 2)

    def test_rule_option3(self):
        """Test rule option"""
        model = self.create_model()
        model.y = Var(initialize=2)
        def f(model):
            yield model.y
            yield 2*model.y
            yield 2*model.y
            yield ObjectiveList.End
        model.c = ObjectiveList(rule=f)
        self.assertEqual(len(model.c), 3)
        self.assertEqual(model.c[1](), 2)
        model.d = ObjectiveList(rule=f(model))
        self.assertEqual(len(model.d), 3)
        self.assertEqual(model.d[1](), 2)

    def test_rule_option4(self):
        """Test rule option"""
        model = self.create_model()
        model.y = Var(initialize=2)
        model.c = ObjectiveList(rule=((i+1)*model.y for i in xrange(3)))
        self.assertEqual(len(model.c), 3)
        self.assertEqual(model.c[1](), 2)

    def test_dim(self):
        """Test dim method"""
        model = self.create_model()
        model.o = ObjectiveList(noruleinit=True)

        self.assertEqual(model.o.dim(),1)

    def test_keys(self):
        """Test keys method"""
        model = self.create_model()
        model.o = ObjectiveList(noruleinit=True)

        self.assertEqual(len(model.o.keys()),0)

    def test_len(self):
        """Test len method"""
        model = self.create_model()
        model.o = ObjectiveList(noruleinit=True)

        self.assertEqual(len(model.o),0)


class MiscObjTests(unittest.TestCase):

    def test_constructor(self):
        a = Objective(name="b")
        self.assertEqual(a.name,"b")
        try:
            a = Objective(foo="bar")
            self.fail("Can't specify an unexpected constructor option")
        except ValueError:
            pass

    def test_contains(self):
        model = ConcreteModel()
        model.a = Set(initialize=[1,2,3])
        model.x = Var()
        def b_rule(model, i):
            return model.x
        model.b = Objective(model.a, rule=b_rule)
 
        self.assertEqual(2 in model.b,True)
        tmp=[]
        for i in model.b:
            tmp.append(i)
        self.assertEqual(len(tmp),3)

    def test_set_get(self):
        a = Objective()
        a.construct()
        try:
            a()
            self.fail("Expect exception because no objective rule was provided.")
        except ValueError:
            pass
        self.assertEqual(a(exception=False),None)
        model = ConcreteModel()
        model.x = Var(initialize=1)
        model.y = Var(initialize=2)
        model.obj = Objective(noruleinit=True)
        model.obj.expr = model.x+model.y
        
        self.assertEqual(model.obj(),3)
        model.x.value = None
        model.y.value = None
        self.assertEqual(model.obj(exception=False),None)
        model.reset()
        self.assertEqual(model.obj(),3)

    def test_rule(self):
        def rule1(model):
            return []
        model = ConcreteModel()
        try:
            model.o = Objective(rule=rule1)
            self.fail("Error generating objective")
        except Exception:
            pass
        #
        def rule1(model):
            return 1.1
        model = ConcreteModel()
        model.o = Objective(rule=rule1)
        self.assertEqual(model.o(),1.1)
        #
        def rule1(model, i):
            return 1.1
        model = ConcreteModel()
        model.a = Set(initialize=[1,2,3])
        try:
            model.o = Objective(model.a,rule=rule1)
        except Exception:
            self.fail("Error generating objective")


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

try:
    unicode
except:
    basestring = str
    long = int

import os.path
import re
import sys
import shutil
from decimal import Decimal

from coopr.core.plugin import alias

from coopr.pyomo.data.TableData import TableData


# format=
# using=
# query=
# user=
# password=
# table=

class db_Table(TableData):

    def __init__(self):
        TableData.__init__(self)
        self.using = None

    def open(self):
        if self.filename is None:
            raise IOError("No data source name was specified")
        if self.filename[0] == '"':
            self.filename = self.filename[1:-1]
        #
        # Initialize self.db
        #
        if not self.using is None:
            self.options.using = self.using
        self.db = None
        if self._data is not None:
            self.db = self._data
        else:
            try:
                self.db = self.connect(self.filename, self.options)
            except Exception:
                raise

    def read(self):
        #
        # Get the table from the database
        #
        if self.db is None:
            return
        cursor = self.db.cursor()
        tmp = []
        if self.options.query is None:
            if self.options.table is None:
                raise IOError("Must specify 'query' or 'table' option!")
            self.options.query = '"SELECT * FROM %s"' % self.options.table
        if not self.options.table is None:
            for row in cursor.columns(table=self.options.table):
                tmp.append(row.column_name)
            tmp=[tmp]
        else:
            # TODO: extend this logic to create a header row for a SQL query
            # FIXME: the current regex logic is pretty brittle...
            match = re.search("SELECT (.*) FROM.*", self.options.query)
            if match is not None:
                fieldstr = match.group(1)
                if fieldstr == "*":
                    raise ValueError("Couldn't extract field names from query. Please specify database columns explicitly in query.")
                else:
                    fields = [f.strip() for f in fieldstr.split(",")]
                #print "FOUND FIELDS", fields # XXX
                tmp = [fields]
                #self.options.index_name = fields[0]
                #self.options.select = fields[1:]
            else:
                # TODO couldn't figure out field names from query; need another strategy!
                pass
        #print "DATA start",self.options.query # XXX
        cursor.execute(self.options.query[1:-1])
        try:
            rows = cursor.fetchall()
        except:
            import logging
            logging.getLogger('coopr.pyomo').error(
                """Fatal error reading from an external ODBC data source.
                
This error was generated outside Pyomo by the Python connector to the
external data source:

    %s

for the query:

    %s

It is possible that you have an error in your external data file,
the ODBC connector for this data source is not correctly installed,
or that there is a bug in the ODBC connector.
""" % (self.filename, self.options.query[1:-1]) )
            raise
        for row in rows:
            #print "DATA",list(row) # XXX
            ttmp=[]
            for data in list(row):
                if isinstance(data,Decimal):
                    ttmp.append(float(data))
                elif data is None:
                    ttmp.append('.')
                elif isinstance(data, str) or isinstance(data, basestring):
                    nulidx = data.find('\x00')
                    if nulidx > -1:
                        data = data[:nulidx]
                    ttmp.append(data)
                else:
                    ttmp.append(data)
            tmp.append(ttmp)
        #print 'FINAL',tmp # XXX
        #
        # Process data from the table
        #
        if type(tmp) in (int,long,float):
            if not self.options.param is None:
                self._info = ["param",self.options.param.name,":=",tmp]
            elif len(self.options.symbol_map) == 1:
                self._info = ["param",self.options.symbol_map[self.options.symbol_map.keys()[0]],":=",tmp]
            else:
                raise IOError("Data looks like a scalar parameter, but multiple parameter names have been specified: %s" % str(self.options.symbol_map))
        elif len(tmp) == 0:
            raise IOError("Empty range '%s'" % self.options.range)
        else:
            #print "SETTING DATA", tmp[0], tmp[1:] # XXX
            #print "OPTIONS", self.options
            self._set_data(tmp[0], tmp[1:])
        return True

    def close(self):
        if self._data is None and not self.db is None:
            del self.db

    def connect(self, connection, options, kwds={}):
        try:
            mod = __import__(options.using)
            args = [connection]
            if not options.user is None:
                args.append(options.user)
            if not options.password is None:
                args.append(options.password)
            return mod.connect(*args, **kwds)
        except ImportError:
            return None


try:
    import pyodbc

    class pyodbc_db_Table(db_Table):

        alias('pyodbc', "Manage IO with a %s database interface" % 'pyodbc')

        _drivers = {  'mdb' : "Microsoft Access Driver (*.mdb)",
                      'xls' : "Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)",
                     'xlsx' : "Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)",
                     'xlsm' : "Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)",
                     'xlsb' : "Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)",
                    'mysql' : "MySQL" }


        def __init__(self):
            db_Table.__init__(self)
            self.using = 'pyodbc'

        def connect(self, connection, options):
            if not options.driver is None:
                ctype = options.driver
            elif '.' in connection:
                ctype = connection.split('.')[-1]
            elif 'mysql' in connection.lower():
                ctype = 'mysql'
            else:
                ctype = ''
            extras = {}
            if ctype in ['xls', 'xlsx', 'xlsm', 'xlsb', 'excel'] or '.xls' in connection or '.xlsx' in connection or '.xlsm' in connection or '.xlsb' in connection:
                extras['autocommit'] = True

            connection = self.create_connection_string(ctype, connection, options)
            try:
                conn = db_Table.connect(self, connection, options, extras)
            except TypeError:
                raise 
            except Exception:
                code, reason = sys.exc_info()[1]
                if code == 'IM002' or code == '08001':
                    # Need a DSN! Try to add it to $HOME/.odbc.ini...
                    odbcIniPath = os.path.join(os.environ['HOME'], '.odbc.ini')
                    if os.path.exists(odbcIniPath):
                        shutil.copy(odbcIniPath, odbcIniPath + '.orig')
                        config = ODBCConfig(filename=odbcIniPath)
                    else:
                        config = ODBCConfig()

                    dsninfo = self.create_dsn_dict(connection, config)
                    dsnid = re.sub('[^A-Za-z0-9]', '', dsninfo['Database']) # Strip filenames of funny characters
                    dsn = 'PYOMO{0}'.format(dsnid)

                    config.add_source(dsn, dsninfo['Driver'])
                    config.add_source_spec(dsn, dsninfo)
                    config.write(odbcIniPath)

                    connstr = "DRIVER={{{0}}};DSN={1}".format(dsninfo['Driver'], dsn)
                    conn = db_Table.connect(self, connstr, options, extras) # Will raise its own exception on failure

                # Propagate the exception
                else:
                    raise Exception((code, reason))

            return conn

        def create_dsn_dict(self, argstr, existing_config):
            result = {}

            parts = argstr.split(';')
            argdict = {}
            for part in parts:
                if len(part) > 0 and '=' in part:
                    key, val = part.split('=',1)
                    argdict[key.lower().strip()] = val.strip()

            if 'driver' in argdict:
                result['Driver'] = "{0}".format(argdict['driver']).strip("{}")

            if 'dsn' in argdict:
                if argdict['dsn'] in existing_config.source_specs:
                    return existing_config.source_specs[argdict['dsn']]
                else:
                    import logging
                    logger = logging.getLogger("coopr.pyomo")
                    logger.warning("DSN with name {0} not found. Attempting to continue with options...".format(argdict['dsn']))

            if 'dbq' in argdict:
                # Using a file for db access.
                if 'Driver' not in result:
                    result['Driver'] = self._drivers[argdict['dbq'].split('.')[-1].lower()]
                result['Database'] = argdict['dbq']
                result['Server'] = 'localhost'
                result['User'] = ''
                result['Password'] = ''
                result['Port'] = '5432'
                result['Description'] = argdict['dbq']
                for k in argdict.keys():
                    if k.capitalize() not in result:
                        result[k.capitalize()] = argdict[k]
            else:
                if 'Driver' not in result:
                    raise Exception("No driver specified, and no DBQ to infer from")
                elif result['Driver'].lower() == "mysql":
                    result['Driver'] = "MySQL"
                    result['Server'] = argdict.get('server', 'localhost')
                    result['Database'] = argdict.get('database', '')
                    result['Port'] = argdict.get('port', '3306')
                    result['Socket'] = argdict.get('socket', '')
                    result['Option'] = argdict.get('option', '')
                    result['Stmt'] = argdict.get('stmt', '')
                    result['User'] = argdict.get('user', '')
                    result['Password'] = argdict.get('password', '')
                    result['Description'] = argdict.get('description', '')
                else:
                    raise Exception("Unknown driver type '{0}' for database connection".format(result['Driver']))

            return result

        def create_connection_string(self, ctype, connection, options):
            if ctype in ['xls', 'xlsx', 'xlsm', 'xlsb', 'excel']:
                return "DRIVER={Microsoft Excel Driver (*.xls, *.xlsx, *.xlsm, *.xlsb)}; Dbq=%s;" % connection
            if ctype in ['mdb', 'access']:
                return "DRIVER={Microsoft Access Driver (*.mdb)}; Dbq=%s;" % connection
            return connection

    class ODBCError(Exception):
        def __init__(self, value):
            self.parameter = value
        def __repr__(self):
            return repr(self.parameter)
    
    class ODBCConfig():
        """
        Encapsulates an ODBC configuration file, usually odbc.ini or
        .odbc.ini, as specified by IBM. ODBC config data can be loaded
        either from a file or a string containing the relevant formatted
        data. Calling load() after initialization will update existing
        information in the config object with the new information.
        """

        def __init__(self, filename=None, data=None):
            """
            Create a new ODBC config instance, loading data from
            the given file and/or data string. Once initialized, the
            new config will contain the data represented in both
            arguments, if any. Data specified as a string argument
            will override that in the file.
            """

            # ugh hardcoded strings. See following URL for info:
            # http://publib.boulder.ibm.com/infocenter/idshelp/v10/index.jsp?topic=/com.ibm.odbc.doc/odbc58.htm
            self.ODBC_DS_KEY = 'ODBC Data Sources'
            self.ODBC_INFO_KEY = 'ODBC'

            self.file = filename

            self.sources = {}
            self.source_specs = {}
            self.odbc_info = {}

            self.load(self.file, data)

        def load(self, filename=None, data=None):
            """
            Load data from the given file and/or data string. If
            both are given, data contained in the string will override
            that in the file. If this config object already contains
            data, the new information loaded will update the old,
            replacing where keys are the same.
            """

            sections = {}
            if filename is not None:
                with open(filename, 'r') as fileHandle:
                    fileData = fileHandle.read()
                    sections.update(self._get_sections(fileData))
            if data is not None:
                sections.update(self._get_sections(data))
            
            if self.ODBC_DS_KEY in sections.keys():
                self.sources.update(sections[self.ODBC_DS_KEY])
                del sections[self.ODBC_DS_KEY]

            if self.ODBC_INFO_KEY in sections.keys():
                self.odbc_info.update(sections[self.ODBC_INFO_KEY])
                del sections[self.ODBC_INFO_KEY]

            self.source_specs.update(sections)

        def __str__(self):
            return "<ODBC config: {0} sources, {1} source specs>".format(len(self.sources), len(self.source_specs))

        def __eq__(self, other):
            if isinstance(other, ODBCConfig):
                return self.sources == other.sources and self.source_specs == other.source_specs and self.odbc_info == other.odbc_info
            return False

        def odbc_repr(self):
            """
            Get the full, odbc.ini-style representation of this
            ODBC configuration.
            """

            str = "[{0}]\n".format(self.ODBC_DS_KEY)
            
            for name in self.sources:
                str += "{0} = {1}\n".format(name, self.sources[name])

            for name in self.source_specs:
                str += "\n[{0}]\n".format(name)
                for key in self.source_specs[name]:
                    str += "{0} = {1}\n".format(key, self.source_specs[name][key])

            if len(self.odbc_info) > 0:
                str += "\n[{0}]\n".format(self.ODBC_INFO_KEY)
                for key in self.odbc_info:
                    str += "{0} = {1}\n".format(key, self.odbc_info[key])

            return str

        def write(self, filename):
            """
            Write the current ODBC configuration to the given file.
            Depends on the odbc_repr() function for a string
            representation of the stored ODBC config information.
            """
            with open(filename, 'w') as f:
                f.write(self.odbc_repr())

        def add_source(self, name, driver):
            """
            Add an ODBC data source to the configuration. A data
            source consists of a unique source name and a driver
            string, which specifies how the source will be loaded.
            If a source name is not unique, it will replace the
            existing source of the same name. A source is required
            in order to have a source specification.
            """

            if name is None or driver is None or len(name) == 0 or len(driver) == 0:
                raise ODBCError("A source must specify both a name and a driver string")
            if name == self.ODBC_DS_KEY or name == self.ODBC_INFO_KEY:
                raise ODBCError("A source cannot use the reserved name '{0}'".format(name))
            self.sources[str(name)] = str(driver)

        def del_source(self, name):
            """
            Remove an ODBC data source from the configuration. If
            any source specifications are based on this source, they
            will be removed as well.
            """

            if name in self.sources.keys():
                if name in self.source_specs.keys():
                    del self.source_specs[name]
                del self.sources[name]

        def add_source_spec(self, name, spec):
            """
            Add an ODBC data source specification to the configuration.
            A source specification consists of a unique name and
            a key-value mapping (i.e. dictionary) of options. In order
            to add a source specification, a data source with a matching
            name must exist in the configuration.
            """

            if name is None or spec is None or len(name) == 0:
                raise ODBCError("A source spec must specify both a name and a spec dictionary")
            if name not in self.sources.keys():
                raise ODBCError("A source spec must have a corresponding source; call .add_source() first")

            self.source_specs[name] = dict(spec)
        
        def del_source_spec(self, name):
            """
            Remove an ODBC data source specification from the
            configuration.
            """

            if name in self.source_specs.keys():
                del self.source_specs[name]

        def set_odbc_info(self, key, value):
            """
            Set an option for the ODBC handling specified in the
            configuration. An option consists of a key-value pair.
            Specifying an existing key will update the current value.
            """

            if key is None or value is None or len(key) == 0 or len(value) == 0:
                raise ODBCError("An ODBC info pair must specify both a key and a value")

            self.odbc_info[str(key)] = str(value)

        def _get_sections(self, data):
            """
            Parse a string for ODBC sections. The parsing algorithm proceeds
            roughly as follows:

            1. Split the string on newline ('\\n') characters.
            2. Remove lines consisting purely of whitespace.
            3. Iterate over lines, storing all key-value pair lines in a dictionary.
            4. When reaching a new section header (denoted by '[str]'), store the old
               key-value pairs under the old section name. Continue from step 3.
            5. On reaching end of data, store the last section and return a mapping
               from section names to dictionaries of key-value pairs in those sections.
            """

            sections = {}
            sectionKey = None
            sectionContents = {}

            emptyLine = re.compile('^[ \t\r]*$')

            lines = data.split('\n')
            for line in lines:
                if emptyLine.match(line):
                    # Whitespace only
                    pass
                elif len(line) < 2:
                    # Not enough room for even 'k='; can't contain info
                    raise ODBCError("Malformed line in ODBC config (no meaningful data): " + line)
                elif line[0] == '[' and line[-1] == ']':
                    # Starts a new section; '=' has no special meaning here
                    sections[sectionKey] = sectionContents
                    sectionKey = line[1:-1]
                    sectionContents = {}
                else:
                    # Not whitespace or section header; must be key=value. No duplicate '=' permitted.
                    if '=' not in line:
                        raise ODBCError("Malformed line in ODBC config (no key-value mapping): " + line)

                    key, value = line.split("=",1)
                    if '=' in value:
                        raise ODBCError("Malformed line in ODBC config (too many '='): " + line)

                    sectionContents[key.strip()] = value.strip()
            sections[sectionKey] = sectionContents

            if None in sections.keys():
                del sections[None]

            return sections

except ImportError:
    pass

try:
    import sqlite3

    class sqlite3_db_Table(db_Table):
        alias('sqlite3', "Manage IO with a %s database interface" % 'sqlite3')

        def __init__(self):
            db_Table.__init__(self)
            self.using = 'sqlite3'

        def connect(self, connection, options):
            assert(options['using'] == 'sqlite3')
            
            filename = connection
            if not os.path.exists(filename):
                raise Exception("No such file: " + filename)

            return sqlite3.connect(filename)

except ImportError:
    pass

for module in ['pymysql']:
    try:
        __import__(module)
        class tmp(db_Table):
            alias(module, "Manage IO with a %s database interface" % module)
            def __init__(self):
                db_Table.__init__(self)
                self.using = module

    except ImportError:
        pass



# A piewise approximiation of a nonconvex objective function.

from coopr.pyomo import *

# Define the function
# Just like in Pyomo constraint rules, a Pyomo model object
# must be the first argument for the function rule
def f(model,t1,t2,x):
    return 0.1*x - cos(5.0*x)


model = ConcreteModel()

# Note we can use an arbitrary number of index sets of 
# arbitrary dimension as the first arguments to the
# Piecewise component.
model.INDEX1 = Set(dimen=2, initialize=[(0,1),(8,3)])
model.X = Var(model.INDEX1, bounds=(-2,2))
model.Z = Var(model.INDEX1)

# For indexed variables, pw_pts must be a
# python dictionary with keys the same as the variable index
PW_PTS = {}

# Increase n to see the solution approach:
# Z[i]=1.19, X[i]=1.89, obj=7.13
n = 3
# Using piecewise representations with a logarithmic number of
# binary variables ('LOG', 'DLOG') requires that pw_pts lists
# must have 2^n + 1 breakpoints.
num_points = 1 + 2**n
step = (2.0 - (-2.0))/(num_points-1)
for idx in model.X.index_set():
    PW_PTS[idx] = [-2.0 + i*step for i in range(num_points)]   # [-2.0, ..., 2.0]

model.linearized_constraint = Piecewise(model.INDEX1,        # indexing sets
                                        model.Z,model.X,     # range and domain variables
                                        pw_pts=PW_PTS,
                                        pw_constr_type='EQ',
                                        pw_repn='LOG',
                                        f_rule=f,
                                        force_pw=True)

# maximize the sum of Z over its index
# This is just a simple example of how to implement indexed variables. All indices
# of Z will have the same solution.
model.obj = Objective(expr= summation(model.Z) , sense=maximize)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Expression', '_ExpressionData']

import sys
import types
import logging
from six import iteritems, iterkeys, next

from coopr.pyomo.base.component import ComponentData, register_component
from coopr.pyomo.base.sparse_indexed_component import SparseIndexedComponent, normalize_index, UnindexedComponent_set
from coopr.pyomo.base.misc import apply_indexed_rule, apply_parameterized_indexed_rule, \
     create_name, tabular_writer
from coopr.pyomo.base.numvalue import NumericConstant, NumericValue, native_types, value, as_numeric
import coopr.pyomo.base.expr

logger = logging.getLogger('coopr.pyomo')

class _ExpressionData(ComponentData, NumericValue):
    """An object that defines an expression that is never cloned"""

    __slots__ = ('_value',)

    # We make value a property so we can ensure that the value
    # assigned to this component is numeric and that all uses of 
    # .value on the NumericValue base class will work
    @property
    def value(self):
        return self._value
    @value.setter
    def value(self, value):
        if value is not None:
            self._value = as_numeric(value)
        else:
            self._value = None

    # TODO: Remove
    def assign(self, value):
        self.value = value

    def __init__(self, owner, value):
        """Constructor"""
        ComponentData.__init__(self, owner)
        self.value = value

    def __getstate__(self):
        state = super(_ExpressionData, self).__getstate__()
        for i in _ExpressionData.__slots__:
            state[i] = getattr(self, i)
        return state

    # Note: because NONE of the slots on this class need to be edited,
    # we don't need to implement a specialized __setstate__ method, and
    # can quietly rely on the super() class's implementation.
    # def __setstate__(self, state):
    #     pass

    def is_constant(self):
        # The underlying expression can always be changed
        # so this should never evaluate as constant
        return False

    def is_fixed(self):        
        return self._value.is_fixed()

    def __call__(self, exception=True):
        """Return the value of this object."""
        if self._value is None:
            return None
        return self._value(exception=exception)

    #################
    # Methods which mock _ExpressionBase behavior defined below
    #################

    # Note: using False here COULD potentially improve performance
    #       inside expression generation and means we wouldn't need to
    #       define a (no-op) clone method. However, there are TONS of
    #       places throughout the code base where is_expression is
    #       used to check whether one needs to "dive deeper" into the
    #       _args
    def is_expression(self):
        return True

    @property
    def _args(self):
        return (self._value,)

    def clone(self):
        return self

    def polynomial_degree(self):
        return self._value.polynomial_degree()

    def to_string(self, ostream=None, verbose=None, precedence=0):
        if ostream is None:
            ostream = sys.stdout
        _verbose = coopr.pyomo.base.expr.TO_STRING_VERBOSE if \
            verbose is None else verbose
        if _verbose:
            ostream.write(str(self))
            ostream.write("{")
        if self._value is None:
            ostream.write("Undefined")
        else:# self._value.is_expression():
            self._value.to_string( ostream=ostream, verbose=verbose, 
                                   precedence=precedence ) 
        #else:
        #    ostream.write(str(self._value))
        if _verbose:
            ostream.write("}")


class Expression(SparseIndexedComponent):
    """A shared expression container, which may be defined over a index"""

    """ Constructor
        Arguments:
           name        The name of this expression
           index       The index set that defines the distinct expression.
                         By default, this is None, indicating that there
                         is a single expression.
           default     A scalar, rule, or dictionary that defines default 
                         values for this expression
           initialize  A dictionary or rule for setting up this expression 
                         with existing model data
    """

    def __new__(cls, *args, **kwds):
        if cls != Expression:
            return super(Expression, cls).__new__(cls)
        if args == ():
            return SimpleExpression.__new__(SimpleExpression)
        else:
            return IndexedExpression.__new__(IndexedExpression)

    def __init__(self, *args, **kwd):
        self._rule       = kwd.pop('initialize', None )
        self._rule       = kwd.pop('expr', self._rule )
        self._rule       = kwd.pop('rule', self._rule)

        self._default_val = None
        default = kwd.pop('default', None )
        
        kwd.setdefault('ctype', Expression)
        SparseIndexedComponent.__init__(self, *args, **kwd)

        self.set_default(default)

        # Because we want to defer the check for defined values until
        # runtime, we will undo the weakref to ourselves
        if not self.is_indexed():
            del self._data[None]

    def _pprint(self):
        return (
            [('Size', len(self)),
             ('Index', None if self._index is UnindexedComponent_set \
                  else self._index),
             ("Default", "(function)" if type(self._default_val) \
                  is types.FunctionType else self._default_val),
             ],
            self.sparse_iteritems(),
            ("Key","Expression"),
            lambda k,v: [ k, 
                          "Undefined" if v.value is None else v.value
                          ]
            )

    def display(self, prefix="", ostream=None):
        """TODO"""
        if not self.active:
            return
        if ostream is None:
            ostream = sys.stdout
        tab="    "
        ostream.write(prefix+self.cname()+" : ")
        ostream.write("Size="+str(len(self)))

        ostream.write("\n")
        tabular_writer( ostream, prefix+tab,
                        ((k,v) for k,v in iteritems(self._data)),
                        ( "Key","Value" ),
                        lambda k, v: [ k,
                                       "Undefined" if v.value is None else value(v.value),
                                       ] )

    def Xpprint(self, ostream=None, verbose=None, precedence=0):
        if ostream is None:
            ostream = sys.stdout
        ostream.write("   %s : " % (self.cname(True),))
        if self.doc is not None:
            ostream.write(" %s\n   " % (self.doc,))
        ostream.write("\tSize=%s\n"
                       % (len(self)) )
        if not self._constructed:
            ostream.write("\tNot constructed\n")
        elif None in self.keys():
            if None in self._data:
                if self[None].value is None:
                    ostream.write("\t%s\n" % None)
                elif self[None].value.is_expression():
                    ostream.write("\t")
                    self[None].value.pprint(ostream=ostream)
                else:
                    ostream.write("\t")
                    self[None].value.pprint(ostream=ostream)
                    ostream.write("\n")
            else:
                ostream.write("\tUndefined\n")
        else:
            for key, val in sorted(self.sparse_iteritems()):
                if val.value is None:
                    ostream.write("\t%s : %s\n" % (key, None))
                elif val.value.is_expression():
                    ostream.write("\t%s : " % key)
                    val.value.pprint(ostream=ostream)
                else:
                    ostream.write("\t%s : " % key)
                    val.value.pprint(ostream=ostream)
                    ostream.write("\n")                    

        if self._default_val is not None:
            ostream.write("\tdefault: ")
            if type(self._default_val) is types.FunctionType:
                ostream.write("(function)\n")
            else:
                ostream.write("%s\n" % self._default_val)

    def __len__(self):
        if self._default_val is None:
            return len(self._data)
        return len(self._index)

    def __contains__(self, ndx):
        if self._default_val is None:
            return ndx in self._data
        return ndx in self._index

    def __iter__(self):
        if self._default_val is None:
            return self._data.__iter__()
        return self._index.__iter__()

    #
    # NB: These are "sparse equivalent" access / iteration methods that
    # only loop over the defined data.
    #

    def sparse_keys(self):
        return list(iterkeys(self._data))

    def sparse_values(self):
        return [ self[x] for x in self._data ]

    def sparse_items(self):
        return [ (x, self[x]) for x in self._data ]

    def sparse_iterkeys(self):
        return iterkeys(self._data)

    def sparse_itervalues(self):
        for key in self._data:
            yield self[key]

    def sparse_iteritems(self):
        if not self.is_indexed():
            # this is hackish, and could be improved.
            for key in self._data:
                yield (key, self[key])
        else:
            # when iterating over sparse values, access the key-value
            # map directly. by definition, the indices are valid, and
            # there is significant overhead associated with the base
            # class __getitem__ method.
            for index, value in iteritems(self._data):
                yield (index, value)

    # TODO: Not sure what "reset" really means in this context...
    def reset(self):
        pass

    #
    # A utility to extract all index-value pairs defining this
    # expression, returned as a dictionary. useful in many contexts,
    # in which key iteration and repeated __getitem__ calls are too
    # expensive to extract the contents of an expression.
    #
    def extract_values(self):
        return dict((key, expression_data.value) \
                    for key, expression_data in iteritems(self))

    #
    # same as above, but only for defined indices.
    #
    def extract_values_sparse(self):
        if not self.is_indexed():
            return dict((key, expression_data.value) \
                        for key, expression_data in self.sparse_iteritems())
        else:
            return dict((key, expression_data.value) \
                        for key, expression_data in iteritems(self._data))
            
    #
    # takes as input a (index, value) dictionary for updating this
    # Expression.  if check=True, then both the index and value are
    # checked through the __getitem__ method of this class.
    #
    def store_values(self, new_values):

        if (self.is_indexed() is False) and (not None in new_values):
            raise RuntimeError("Cannot store value for scalar Expression"
                               "="+self.cname(True)+"; no value with index "
                               "None in input new values map.")

        if self.is_indexed():
            for index, new_value in iteritems(new_values):
                self._data[index].value = new_value
        else:
            # scalars have to be handled differently
            self[None].value = new_values[None]

    def _default(self, idx):

        if not self._constructed:
            if idx is None:
                idx_str = '%s' % (self.cname(True),)
            else:
                idx_str = '%s[%s]' % (self.cname(True), idx,)
            raise ValueError(
                "Error retrieving Expression value (%s): The Expression value "
                "has not been constructed" % ( idx_str,) )                
 
        if self._default_val is None:
            if idx is None:
                idx_str = '%s' % (self.cname(True),)
            else:
                idx_str = '%s[%s]' % (self.cname(True), idx,)
            raise KeyError(
                "Error retrieving Expression value (%s): The Expression value "
                "is undefined and no default value is specified"
                % ( idx_str,) )

        _default_type = type(self._default_val)
        if _default_type in native_types:
            val = self._default_val
        elif _default_type is types.FunctionType:
            val = apply_indexed_rule(
                self, self._default_val, self.parent(), idx )
        elif hasattr(self._default_val, '__getitem__'):
            try:
                val = self._default_val[idx]
            except IndexError:
                val = self._default_val
        else:
            val = self._default_val

        self[idx] = val
        return val

    def set_default(self, val):
        self._default_val = val

    # the __setitem__ method below performs significant validation around
    # the input indices, and processing / conversion of values. in various
    # contexts, we don't need to incur this overhead, specifically during
    # initialization. assumes the input value is in the set native_types
    def _raw_setitem(self, ndx, val):

        if ndx is None:
            self.value = val
            self._data[ndx] = None
        elif ndx in self._data:
            self._data[ndx].value = val
        else:
            self._data[ndx] = _ExpressionData(self, val)

    def __setitem__(self, ndx, val):
        

        if ndx not in self._index:

            # We rely (for performance purposes) on "most" people doing things 
            # correctly; that is, they send us either a scalar or a valid tuple.  
            # So, for efficiency, we will check the index *first*, and only go
            # through the hassle of flattening things if the ndx is not found.
            ndx = normalize_index(ndx)
            if ndx not in self._index:
                if not self.is_indexed():
                    msg = "Error setting Expression value: " \
                          "Cannot treat the scalar Expression '%s' as an array" \
                          % ( self.cname(True), )
                else:
                    msg = "Error setting Expression value: " \
                          "Index '%s' is not valid for array Expression '%s'" \
                          % ( ndx, self.cname(True), )
                raise KeyError(msg)

        # we have a valid index and value - do the actual set.
        self._raw_setitem(ndx, val)

    def construct(self, data=None):
        """ Apply the rule to construct values in this set """

        if __debug__:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Constructing Expression, name=%s, from data=%s"
                             % ( self.cname(True), str(data) ))

        if self._constructed:
            raise IOError(
                "Cannot reconstruct Expression '%s' (already constructed)"
                % self.cname(True) )
            return

        #
        # Construct using the initial data or the data loaded from an
        # external source. Data values will be queried in to following
        # order:
        #   1) the 'data' dictionary
        #   2) the self._rule dictionary or rule
        #   3) [implicit: fall back on the default value]
        #
        # To accomplish this, we will first set all the values based on
        # self._rule, and then allow the data dictionary to
        # overwrite anything.
        #
        # NB: Scalar Expressions can always be treated as "normal" indexed
        # expression, indexed by a set {None}.
        #
        #
        # NB: Previously, we would raise an exception for constructing
        # scalar Expressions with no defined data.  As that was a special
        # case (i.e. didn't apply to arrays) and was frustrating for
        # Concrete folks who were going to initialize the value later,
        # we will allow an undefined Expression to be constructed and will
        # instead throw an exception later if the user tries to *use* the
        # Expression before it is initialized.
        #
        _init = self._rule
        if _init is not None:
            _init_type = type(_init)
            if _init_type is dict:
                for key, val in iteritems(_init):
                    self[key] = val
            elif _init_type is types.FunctionType:
                if self.is_indexed():
                    # a situation can arise in which the expression is
                    # indexed by an empty set, or an empty set
                    # tuple. rare, but it has happened.
                    if len(self._index) != 0: 
                        _iter = self._index.__iter__()
                        idx = next(_iter)
                        val = apply_indexed_rule(self, _init, self.parent(), idx)
                        if type(val) is dict:
                            for key, v in iteritems(val):
                                self[key] = v
                        else:
                            # the following is optimized, due to its
                            # frequent occurence in practice. failing
                            # to perform these optimizations
                            # unnecessarily blows up run times, on a
                            # number of large-scale models.

                            self_parent = self.parent()

                            for idx in self._index:

                                # IMPORTANT: We assume that the
                                #            iterator over an index
                                #            set returns a flattened
                                #            tuple. Otherwise, the
                                #            validation process is far
                                #            too expensive.

                                val = apply_indexed_rule(self,
                                                         _init,
                                                         self_parent,
                                                         idx)

                                # At this point, we know _init is a
                                # raw native type value, or one that
                                # can be treated as such.  we have a
                                # valid index and value - set it.
                                self._raw_setitem(idx, val)

                else:
                    self[None] = _init(self.parent())
            else:
                if isinstance(_init, SparseIndexedComponent):
                    # Ideally, we want to reduce SparseIndexedComponents
                    # to a dict, but without "densifying" it.  However,
                    # since there is no way to (easily) get the default
                    # value, we will take the "less surprising" route of
                    # letting the source become dense, so that we get
                    # the expected copy.
                    sparse_src = len(_init) != len(_init.keys())
                    tmp = {}
                    for key in _init.keys():
                        try:
                            val = _init[key]
                        except ValueError:
                            continue
                        tmp[key] = val
                    tmp = dict(iteritems(_init))
                    if sparse_src and len(_init) == len(_init.keys()):
                        logger.warning("""
Initializing Expression %s using a sparse mutable indexed component (%s).
    This has resulted in the conversion of the source to dense form.
""" % ( self.cname(True), _init.name ) )
                    _init = tmp

                # if things look like a dictionary, then we will treat
                # it as such
                _isDict = hasattr(_init, '__getitem__')
                if _isDict:
                    try:
                        for x in _init:
                            _init.__getitem__(x)
                    except:
                        _isDict = False
                if _isDict:
                    for key in _init:
                        self[key] = _init[key]
                else:
                    # the following is optimized, due to its frequent occurence
                    # in practice. failing to perform these optimizations
                    # unnecessarily blows up run times, on a number of large-scale
                    # models.
                    for key in self._index:

                        # IMPORTANT: We assume that the iterator over an index
                        #            set returns a flattened tuple. Otherwise,
                        #            the validation process is far too expensive.

                        # At this point, we know _init is a raw native type
                        # value, or one that can be treated as such. 
                        # we have a valid index and value - set it.
                        self._raw_setitem(key, _init)


        if data is not None:
            try:
                for key, val in iteritems(data):
                    self[key] = val
            except Exception:
                msg = sys.exc_info()[1]
                if type(data) not in [dict]:
                   raise ValueError("Attempting to initialize Expression="+self.cname(True)+" with data="+str(data)+". Data type is not a dictionary, and a dictionary is expected. Did you create a dictionary with a \"None\" index?")
                else:
                    raise RuntimeError("Failed to set value for Expression="+self.cname(True)+", index="+str(key)+", value="+str(val)+"; source error message="+str(msg))

        self._constructed = True
        self.set_default(self._default_val)


    #
    # a simple utility to set all of the data elements in the 
    # index set of this expression to the common input value.
    # if the entries are there, it will over-ride the values.
    # if the entries are not there, it will create them.
    #
    def initialize_all_to(self, common_value):

        if self.is_indexed():
            for index in self._index:
                self._data[index] = _ExpressionData(self, common_value)
        else:
            self[None] = common_value


class SimpleExpression(_ExpressionData, Expression):

    def __init__(self, *args, **kwds):
        Expression.__init__(self, *args, **kwds)
        _ExpressionData.__init__(self, self, kwds.get('default',None))

    # Since this class derives from Component and Component.__getstate__
    # just packs up the entire __dict__ into the state dict, there s
    # nothng special that we need to do here.  We will just defer to the
    # super() get/set state.  Since all of our get/set state methods
    # rely on super() to traverse the MRO, this will automatically pick
    # up both the Component and Data base classes.
    #
    #def __getstate__(self):
    #    pass
    #
    #def __setstate__(self, state):
    #    pass

    def Xpprint(self, ostream=None, verbose=None, nested=False, eol_flag=True, precedence=0):
        # Needed so that users find Expression.pprint and not
        # _ExpressionData.pprint
        if precedence == 0:
            Expression.pprint(self, ostream=ostream, verbose=None)
        else:
            ostream.write(str(self))

    def __call__(self, exception=True):

        if self._constructed:
            return _ExpressionData.__call__(self, exception=exception)
        if exception:
            raise ValueError("Evaluating the numeric value of expression '%s' "
                             "before the Expression has been constructed (there "
                             "is currently no value to return)."
                             % self.cname(True))

    def set_value(self, value):
        self[None] = value
        

class IndexedExpression(Expression):

    def __call__(self, exception=True):
        """Compute the value of the expression"""
        if exception:
            msg = 'Cannot compute the value of an array of expressions'
            raise TypeError(msg)


register_component(Expression,
                   "Named expressions that can be used in other expressions.")




# A Suffix example for the collection of duals.

from coopr.pyomo import *

### Create a trivial and infeasible example model
model = ConcreteModel()
model.x = Var(within=NonNegativeReals)
model.obj = Objective(expr=model.x)
model.con = Constraint(expr=model.x >= 1)
###

# Declare an IMPORT Suffix to store the dual information that will
# be returned by the solver. When Suffix components are declared 
# with an IMPORT direction, Pyomo solver interfaces will attempt to collect
# this named information from a solver solution.
model.dual = Suffix(direction=Suffix.IMPORT)




#
# Unit Tests for Utility Functions
#

import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep

import pyutilib.th as unittest
from coopr.pyomo import *
from nose.tools import nottest
import pickle

from six import StringIO

def obj_rule(model):
    return sum(model.x[a] + model.y[a] for a in model.A)
def constr_rule(model,a):
    return model.x[a] >= model.y[a]

class Test(unittest.TestCase):

    def test_expr1(self):
        model = AbstractModel()
        model.A = Set(initialize=[1,2,3])
        model.B = Param(model.A,initialize={1:100,2:200,3:300}, mutable=True)
        model.x = Var(model.A)
        model.y = Var(model.A)
        instance=model.create()
        expr = dot_product(instance.x,instance.B,instance.y)
        self.assertEquals(
            str(expr),
            "x[1] * B[1] * y[1] + x[2] * B[2] * y[2] + x[3] * B[3] * y[3]" )

    def test_expr2(self):
        model = AbstractModel()
        model.A = Set(initialize=[1,2,3])
        model.B = Param(model.A,initialize={1:100,2:200,3:300}, mutable=True)
        model.x = Var(model.A)
        model.y = Var(model.A)
        instance=model.create()
        expr = dot_product(instance.x,instance.B,instance.y, index=[1,3])
        self.assertEquals(
            str(expr),
            "x[1] * B[1] * y[1] + x[3] * B[3] * y[3]" )

    def test_expr3(self):
        model = AbstractModel()
        model.A = Set(initialize=[1,2,3])
        model.B = Param(model.A,initialize={1:100,2:200,3:300}, mutable=True)
        model.x = Var(model.A)
        model.y = Var(model.A)
        instance=model.create()
        expr = dot_product(instance.x,instance.B,denom=instance.y, index=[1,3])
        self.assertEquals(
            str(expr),
            "x[1] * B[1] / y[1] + x[3] * B[3] / y[3]" )

    def test_expr4(self):
        model = AbstractModel()
        model.A = Set(initialize=[1,2,3])
        model.B = Param(model.A,initialize={1:100,2:200,3:300}, mutable=True)
        model.x = Var(model.A)
        model.y = Var(model.A)
        instance=model.create()
        expr = dot_product(denom=[instance.y,instance.x])
        self.assertEquals(
            str(expr),
            "1 / ( y[1] * x[1] ) + 1 / ( y[2] * x[2] ) + 1 / ( y[3] * x[3] )" )

    def test_expr5(self):
        model = ConcreteModel()
        model.A = Set(initialize=[1,2,3], doc='set A')
        model.B = Param(model.A, initialize={1:100,2:200,3:300}, doc='param B', mutable=True)
        model.C = Param(initialize=3, doc='param C', mutable=True)
        model.x = Var(model.A, doc='var x')
        model.y = Var(doc='var y')
        model.o = Objective(expr=model.y, doc='obj o')
        model.c1 = Constraint(expr=model.x[1] >= 0, doc='con c1')
        def c2_rule(model, a):
            return model.B[a] * model.x[a] <= 1
        model.c2 = Constraint(model.A, doc='con c2')
        model.c3 = ConstraintList(doc='con c3')
        model.c3.add(model.y <= 0)
        #
        instance=model.create()
        OUTPUT=open(currdir+"test_expr5.out","w")
        instance.pprint(ostream=OUTPUT)
        OUTPUT.close()
        self.assertFileEqualsBaseline(currdir+"test_expr5.out",currdir+"test_expr5.txt")

if __name__ == "__main__":
    unittest.main()



# Author: Hans Pirnay, 2012-11-27
# This is the Hicks reactor example from the sIPOPT paper
#
# This Pyomo example is formulated as a python script.
# To run this script execute the following command:
#
# $ coopr_python sipopt_hicks.py
#
# Execution of this script requires that the ipopt_sens
# solver (distributed with Ipopt) is in the current search
# path for executables on this system. Optionally required
# are the numpy and matplotlib python modules (needed for 
# viewing results).

import coopr.environ
from coopr.pyomo import *
from coopr.opt import SolverFactory

### Create the ipopt_sens solver plugin using the ASL interface
solver = 'ipopt_sens'
solver_io = 'nl'
stream_solver = False    # True prints solver output to screen
keepfiles =     False    # True prints intermediate file names (.nl,.sol,...)
opt = SolverFactory(solver,solver_io=solver_io)
###

if opt is None:
    print("")
    print("ERROR: Unable to create solver plugin for 'ipopt_sens'")
    print("")
    exit(1)

### model definition
# parameters
nfe = 40
ncp = 3
FE = range(nfe)
CP = range(ncp)

time = 9.0
jj = 100.0
cf = 7.6
alpha = 1.95e-4
tf = 300.0
k10 = 300.0
tc = 290.0
n = 5.0
alpha1 = 1.0e6
alpha2 = 2.0e3
alpha3 = 1.e-3
c_des = 0.0944
t_des = 0.7766
u_des = 340.0
c_init = 0.1367
t_init = 0.7293
u_init = 390.0
theta = 20.0
yc = tc/(jj*cf)
yf = tf/(jj*cf)

model = ConcreteModel()
model.c_init_var = Var()
model.t_init_var = Var()
model.cdot = Var(FE,CP)
model.tdot = Var(FE,CP)
def c_init_rule(m, i, j):
    return float(i)/nfe*(c_des-c_init) + c_init
model.c    = Var(FE,CP, within=NonNegativeReals, initialize=c_init_rule)
def t_init_rule(m, i, j):
    return float(i)/nfe*(t_des-t_init) + t_init
model.t    = Var(FE,CP, within=NonNegativeReals, initialize=t_init_rule)
model.u    = Var(FE, within=NonNegativeReals, initialize=1.0)

a_init = {}
a_init[0,0] = 0.19681547722366
a_init[0,1] = 0.39442431473909
a_init[0,2] = 0.37640306270047
a_init[1,0] = -0.06553542585020
a_init[1,1] = 0.29207341166523
a_init[1,2] = 0.51248582618842
a_init[2,0] = 0.02377097434822
a_init[2,1] = -0.04154875212600
a_init[2,2] = 0.11111111111111

model.a = Param(FE,CP, initialize=a_init)

h = [1.0/nfe]*nfe

def cdot_ode_rule(m, i, j):
    return m.cdot[i,j] == (1.0-m.c[i,j])/theta-k10*exp(-n/m.t[i,j])*m.c[i,j]
model.cdot_ode = Constraint(FE, CP, rule=cdot_ode_rule)

def tdot_ode_rule(m, i, j):
    return m.tdot[i,j] == (yf-m.t[i,j])/theta+k10*exp(-n/m.t[i,j])*m.c[i,j]-alpha*m.u[i]*(m.t[i,j]-yc)
model.tdot_ode = Constraint(FE, CP, rule=tdot_ode_rule)

def fecolc_rule(m, i,j):
    if i==0:
        return m.c[i,j] == m.c_init_var + time*h[i]*sum(m.a[k,j]*m.cdot[i,k] for k in CP)
    else:
        return m.c[i,j] == m.c[i-1,ncp-1] + time*h[i]*sum(m.a[k,j]*m.cdot[i,k] for k in CP)
model.fecolc = Constraint(FE, CP, rule=fecolc_rule)
model.c_init_def = Constraint(expr= model.c_init_var == c_init)
model.t_init_def = Constraint(expr= model.t_init_var == t_init)

def fecolt_rule(m, i, j):
    if i==0:
        return m.t[i,j] == m.t_init_var + time*h[i]*sum(m.a[k,j]*m.tdot[i,k] for k in CP)
    else:
        return m.t[i,j] == m.t[i-1,ncp-1] + time*h[i]*sum(m.a[k,j]*m.tdot[i,k] for k in CP)
model.fecolt = Constraint(FE, CP, rule=fecolt_rule)

def obj_rule(m):
    return \
      sum(h[i]*sum((alpha1*(m.c[i,j]-c_des)**2+ alpha2*(m.t[i,j]-t_des)**2+alpha3*(m.u[i]-u_des)**2 )*m.a[j,ncp-1] for j in CP) for i in range(2,nfe)) + \
      h[0]*sum((alpha1*((m.c_init_var+time*h[0]*sum( m.a[k,j]*m.cdot[0,k] for k in CP)) - c_des)**2 + alpha2*((m.t_init_var+time*h[0]*sum(m.a[k,j]*m.tdot[0,k]for k in CP))-t_des)**2 + alpha3*(m.u[0]-u_des)**2)*m.a[j,ncp-1] for j in CP)
model.cost = Objective(rule=obj_rule)
###

### declare suffixes
model.sens_state_0 = Suffix(direction=Suffix.EXPORT)
model.sens_state_1 = Suffix(direction=Suffix.EXPORT)
model.sens_state_value_1 = Suffix(direction=Suffix.EXPORT)
model.sens_sol_state_1  = Suffix(direction=Suffix.IMPORT)
model.sens_init_constr  = Suffix(direction=Suffix.EXPORT)
###

### set sIPOPT data
opt.options['run_sens'] = 'yes'
model.sens_state_0.setValue(model.c_init_var, 1)
model.sens_state_0.setValue(model.t_init_var, 2)
model.sens_state_1.setValue(model.c[4,0], 1)
model.sens_state_1.setValue(model.t[4,0], 2)
model.sens_state_value_1.setValue(model.c[4,0], 0.135)
model.sens_state_value_1.setValue(model.t[4,0], 0.745)
model.sens_init_constr.setValue(model.c_init_def, 1)
model.sens_init_constr.setValue(model.t_init_def, 1)
###

### Send the model to ipopt_sens and collect the solution
results = opt.solve(model,keepfiles=keepfiles,tee=stream_solver)
# load the results (including any values for previously declared
# IMPORT / IMPORT_EXPORT Suffix components)
model.load(results)
###

# Plot the results
try:
    import numpy as np
    import matplotlib.pyplot as plt
except ImportError:
    print("")
    print("ERROR: numpy and matplotlib are required to view the example results")
    print("")
    exit(1)

def collocation_points(n_fe, n_cp, h):
    t = 0.0
    r1 = 0.15505102572168
    r2 = 0.64494897427832
    r3 = 1.0
    for i in range(n_fe):
            yield t+h[i]*r1
            yield t+h[i]*r2
            yield t+h[i]*r3
            t += h[i]

def collocation_idx(n_fe, n_cp):
    for i in range(n_fe):
        yield i,0
        yield i,1
        yield i,2

times = np.array([i for i in collocation_points(nfe, ncp, h)])
cnominal = np.zeros((nfe*ncp,1))
cperturbed = np.zeros((nfe*ncp,1))
tnominal = np.zeros((nfe*ncp,1))
tperturbed = np.zeros((nfe*ncp,1))
for k,(i,j) in enumerate(collocation_idx(nfe, ncp)):
    cnominal[k] = model.c[i,j]
    tnominal[k] = model.t[i,j]
    cperturbed[k] = model.sens_sol_state_1.getValue(model.c[i,j])
    tperturbed[k] = model.sens_sol_state_1.getValue(model.t[i,j])

plt.subplot(2,1,1)
plt.plot(times, cnominal, label='c_nominal')
plt.hold(True)
plt.plot(times, cperturbed, label='c_perturbed')
plt.xlim([min(times),max(times)])
plt.legend(loc=0)
plt.subplot(2,1,2)
plt.plot(times, tnominal, label='t_nominal')
plt.plot(times, tperturbed, label='t_perturbed')
plt.xlim([min(times),max(times)])
plt.legend(loc=0)
plt.show()



__all__ = ['IPreCheckHook', 'IPostCheckHook']

from coopr.core.plugin import *

class IPreCheckHook(Interface):

    def precheck(self, runner, script, info):
        pass


class IPostCheckHook(Interface):

    def postcheck(self, runner, script, info):
        pass



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['ModelData']

import logging
from coopr.pyomo.base.plugin import *

logger = logging.getLogger('coopr.pyomo')

class ModelData(object):
    """
    An object that manages data for a model.

    This object contains the interface routines for importing and
    exporting data from external data sources.
    """

    def __init__(self, **kwds):
        """
        Constructor
        """
        # maps initialization data as follows: _data[namespace][symbol] -> data
        self._data={}

        # This is the data that is imported from various sources
        self._default={}

        # the model for which this data is associated. used for error
        # checking, e.g., object names must exist in the model, set 
        # dimensions must match, etc.
        self._model = kwds.pop('model', None)

        # a list of objects that implement the IDataManager interface.
        self._info=[]

        # if one desires, one can add a filename (but only one) to the 
        # ModelData object in the constructor call.
        if 'filename' in kwds:
            if not kwds['filename'] is None:
                filename = kwds['filename']
                del kwds['filename']
                self.add(filename, **kwds)
                self.read()
        elif 'data_dict' in kwds:
            if not kwds['data_dict'] is None:
                if self._model is None:
                    raise ValueError("Cannot read model data without a model")
                self._data = kwds['data_dict']
        elif len(kwds) > 0:
            raise ValueError("Unknown options: %s" % str(kwds.keys()))

    #
    # intent of this method is to add, on a component-by-component 
    # basis, initialization data to a ModelData instance. not the
    # "usual" form of data initialization.
    #
    def initialize(self, component, data, namespace=None):
        if namespace not in self._data.keys():
            self._data[namespace]={}
        self._data[namespace][component] = data

    #
    # construct data source objects associated with the input filename,
    # and store them for subsequent processing by this class. important
    # to note that this does not actually *read* the data. this is the
    # more typical use case for data initialization.
    #
    def add(self, filename, **kwds):
        # the data manager factory is keyed off the data file suffix.
        tmp = filename.split(".")[-1]
        data = DataManagerFactory(tmp) 
        if data is None:
            raise IOError("The file extension '%s' is not supported with your current Python installation.  Cannot add data from file '%s'." % (tmp, filename))
        data.initialize(filename=filename, **kwds)
        self._info.append(data)

    #
    # given a set of data sources that have been associated with this
    # ModelData object, read in and validate the data, populating the
    # the _data and _default member dictionaries.
    #
    def read(self, model=None):

        if model is not None:
            self._model=model

        if self._model is None:
            raise ValueError("Cannot read model data without a model")
        #
        # Although we want to load data in a particular order, it
        # may make sense to open/close all data resources all at once.
        # Should we do this by default, or make this a configuration option
        # for this class?
        #
        for data in self._info:
            if type(data.options.index) is tuple:
                data.options.index = list(data.options.index)
            elif not data.options.index is None and not type(data.options.index) is list:
                data.options.index = [ data.options.index ]
            if type(data.options.param) is tuple:
                data.options.param = list(data.options.param)
            elif not data.options.param is None and not type(data.options.param) is list:
                data.options.param = [ data.options.param ]

            if not data.options.index is None:
                data.options.select = data.options.index + data.options.param
                data.options.index = None
            if not data.options.set is None:
                data.options.set = getattr(self._model, data.options.set)
            if not data.options.param is None:
                if not data.options.param_name is None:
                    ans = []
                    for _param in data.options.param:
                        ans.append( getattr(self._model, data.options.param_name[_param]) )
                    data.options.param = ans
                elif type(data.options.param) in (list,tuple):
                    ans = []
                    for i in range(len(data.options.param)):
                        ans.append( getattr(self._model, data.options.param[i]) )
                    data.options.param = ans
                else:
                    data.options.param = getattr(self._model, data.options.param)
            if not data.options.index_name is None:
                data.options.index = getattr(self._model, data.options.index_name)

            if __debug__:
                logger.debug("Importing data...")

            data.open()
            if not data.read():
                print("Warning: error occured while processing %s" % str(data))
            data.close()
            if __debug__:
                logger.debug("Processing data ...")
            #
            # Update the _data and _default dictionaries with data that
            # was read in.
            #
            status = data.process(self._model, self._data, self._default)
            data.clear()
            if __debug__:
                logger.debug("Done.")




#
# Unit Tests for ModelData objects
#

import os
import sys
from os.path import abspath, dirname
coopr_dir=dirname(dirname(abspath(__file__)))+os.sep+".."+os.sep+".."

from coopr.pyomo import *
import coopr
import pyutilib.common
import pyutilib.th as unittest
from coopr.pyomo.base.plugin import DataManagerFactory
import coopr.environ

currdir=dirname(abspath(__file__))+os.sep
example_dir=coopr_dir+os.sep+".."+os.sep+"examples"+os.sep+"pyomo"+os.sep+"tutorials"+os.sep+"tab"+os.sep
csv_dir=coopr_dir+os.sep+".."+os.sep+"examples"+os.sep+"pyomo"+os.sep+"tutorials"+os.sep+"csv"+os.sep
tutorial_dir=coopr_dir+os.sep+".."+os.sep+"examples"+os.sep+"pyomo"+os.sep+"tutorials"+os.sep

xls_interface = not DataManagerFactory('xls') is None
xlsx_interface = not DataManagerFactory('xlsx') is None
xlsb_interface = not DataManagerFactory('xlsb') is None
xlsm_interface = not DataManagerFactory('xlsm') is None


class PyomoTableData(unittest.TestCase):

    def setUp(self):
        pass

    def construct(self,filename):
        pass

    def test_read_set(self):
        td = DataManagerFactory('xls')
        td.initialize(currdir+"Book1.xls", range="TheRange", format='set', set="X")
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual( td._info, ['set', 'X', ':=', 'A1', 2.0, 3.0, 4.0, 'A5', 6.0, 7.0, 8.0, 'A9', 10.0, 11.0, 12.0, 'A13', 14.0, 15.0, 16.0])
        except pyutilib.common.ApplicationError:
            pass

    def test_read_param1(self):
        td = DataManagerFactory('xls')
        td.initialize(currdir+"Book1.xls", range="TheRange", index=['aa'], param=['bb','cc','dd'])
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual( td._info, ['param', ':', 'bb', 'cc', 'dd', ':=', 'A1', 2.0, 3.0, 4.0, 'A5', 6.0, 7.0, 8.0, 'A9', 10.0, 11.0, 12.0, 'A13', 14.0, 15.0, 16.0])
        except pyutilib.common.ApplicationError:
            pass

    def test_read_param2(self):
        td = DataManagerFactory('xls')
        td.initialize(currdir+"Book1.xls",range="TheRange", index_name="X", index=['aa'], param=['bb','cc','dd'])
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual( td._info, ['param', ':', 'X', ':', 'bb', 'cc', 'dd', ':=', 'A1', 2.0, 3.0, 4.0, 'A5', 6.0, 7.0, 8.0, 'A9', 10.0, 11.0, 12.0, 'A13', 14.0, 15.0, 16.0])
        except pyutilib.common.ApplicationError:
            pass

    def test_read_param3(self):
        td = DataManagerFactory('xls')
        td.initialize(currdir+"Book1.xls",range="TheRange", index_name="X", index=['aa','bb','cc'], param=["dd"], param_name={'dd':'a'})
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual( td._info, ['param', ':', 'X', ':', 'a', ':=', 'A1', 2.0, 3.0, 4.0, 'A5', 6.0, 7.0, 8.0, 'A9', 10.0, 11.0, 12.0, 'A13', 14.0, 15.0, 16.0])
        except pyutilib.common.ApplicationError:
            pass

    def test_read_param4(self):
        td = DataManagerFactory('xls')
        td.initialize(currdir+"Book1.xls",range="TheRange", index_name="X", index=['aa','bb'], param=['cc','dd'], param_name={'cc':'a', 'dd':'b'})
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual( td._info, ['param', ':', 'X', ':', 'a', 'b', ':=', 'A1', 2.0, 3.0, 4.0, 'A5', 6.0, 7.0, 8.0, 'A9', 10.0, 11.0, 12.0, 'A13', 14.0, 15.0, 16.0])
        except pyutilib.common.ApplicationError:
            pass

    def test_read_array1(self):
        td = DataManagerFactory('xls')
        td.initialize(currdir+"Book1.xls",range="TheRange", param="X", format="array")
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual( td._info, ['param', 'X', ':', 'bb', 'cc', 'dd', ':=', 'A1', 2.0, 3.0, 4.0, 'A5', 6.0, 7.0, 8.0, 'A9', 10.0, 11.0, 12.0, 'A13', 14.0, 15.0, 16.0])
        except pyutilib.common.ApplicationError:
            pass

    def test_read_array2(self):
        td = DataManagerFactory('xls')
        td.initialize(currdir+"Book1.xls",range="TheRange",param="X",format="transposed_array")
        try:
            td.open()
            td.read()
            td.close()
            self.assertEqual( td._info, ['param', 'X', '(tr)',':', 'bb', 'cc', 'dd', ':=', 'A1', 2.0, 3.0, 4.0, 'A5', 6.0, 7.0, 8.0, 'A9', 10.0, 11.0, 12.0, 'A13', 14.0, 15.0, 16.0])
        except pyutilib.common.ApplicationError:
            pass

    def test_error1(self):
        td = DataManagerFactory('xls')
        td.initialize("bad")
        try:
            td.open()
            self.fail("Expected IOError because of bad file")
        except IOError:
            pass

    def test_error2(self):
        td = DataManagerFactory('xls')
        try:
            td.open()
            self.fail("Expected IOError because no file specified")
        except (IOError,AttributeError):
            pass

    def test_error3(self):
        td = DataManagerFactory('txt')
        try:
            td.initialize(currdir+"display.txt")
            td.open()
            self.fail("Expected IOError because of bad file type")
        except (IOError, AttributeError):
            pass

    def test_error4(self):
        td = DataManagerFactory('txt')
        try:
            td.initialize(filename=currdir+"dummy")
            td.open()
            self.fail("Expected IOError because of bad file type")
        except (IOError, AttributeError):
            pass

    def test_error5(self):
        td = DataManagerFactory('tab')
        td.initialize(example_dir+"D.tab", param="D", format="foo")
        td.open()
        try:
            td.read()
            self.fail("Expected IOError because of bad format")
        except ValueError:
            pass

PyomoTableData = unittest.skipIf(not xls_interface, "No XLS interface available")(PyomoTableData)

class PyomoModelData(unittest.TestCase):

    def test_md1(self):
        md = ModelData()
        md.add(example_dir+"A.tab")
        try:
            md.read()
            self.fail("Must specify a model")
        except ValueError:
            pass
        model=AbstractModel()
        try:
            md.read(model)
            self.fail("Expected ValueError")
        except ValueError:
            pass
        model.A=Set()

    def test_md2(self):
        md = ModelData()
        md.add(currdir+"data1.dat")
        model=AbstractModel()
        model.A=Set()
        md.read(model)

    def test_md3(self):
        md = ModelData()
        md.add(currdir+"data2.dat")
        model=AbstractModel()
        model.A=Set()
        try:
            md.read(model)
            self.fail("Expected error because of extraneous text")
        except IOError:
            pass

    def test_md4(self):
        md = ModelData()
        md.add(currdir+"data3.dat")
        model=AbstractModel()
        model.A=Set()
        model.B=Set()
        model.C=Set()
        md.read(model)

    def test_md5(self):
        md = ModelData()
        md.add(currdir+"data4.dat")
        model=AbstractModel()
        model.A=Set()
        try:
            md.read(model)
        except (ValueError,IOError):
            pass

    def test_md6(self):
        md = ModelData()
        md.add(currdir+"data5.dat")
        model=AbstractModel()
        model.A=Set()
        try:
            md.read(model)
        except ValueError:
            pass

    def test_md7(self):
        md = ModelData()
        md.add(currdir+"data1.tab")
        model=AbstractModel()
        try:
            md.read(model)
            self.fail("Expected IOError")
        except IOError:
            pass

    def test_md8(self):
        md = ModelData()
        md.add(currdir+"data6.dat")
        model=AbstractModel()
        model.A=Set()
        try:
            md.read(model)
            self.fail("Expected IOError")
        except IOError:
            pass

    def test_md9(self):
        md = ModelData()
        md.add(currdir+"data7.dat")
        model=AbstractModel()
        model.A=Set()
        model.B=Param(model.A)
        md.read(model)

    def test_md10(self):
        md = ModelData()
        md.add(currdir+"data8.dat")
        model=AbstractModel()
        model.A=Param(within=Boolean)
        model.B=Param(within=Boolean)
        model.Z=Set()
        md.read(model)
        instance = model.create(md)
        #self.assertEqual(instance.Z.data(), set(['foo[*]' 'bar' '[' '*' ']' 'bar[1,*,a,*]' 'foo-bar' 'hello-goodbye']))

    def test_md11(self):
        cwd = os.getcwd()
        os.chdir(currdir)
        md = ModelData()
        md.add(currdir+"data11.dat")
        model=AbstractModel()
        model.A=Set()
        model.B=Set()
        model.C=Set()
        model.D=Set()
        md.read(model)
        os.chdir(cwd)

    def test_md11(self):
        cwd = os.getcwd()
        os.chdir(currdir)
        model=AbstractModel()
        model.a=Param()
        model.b=Param()
        model.c=Param()
        model.d=Param()
        # Test 1
        instance = model.create(currdir+'data14.dat', namespaces=['ns1','ns2'])
        self.assertEqual( value(instance.a), 1)
        self.assertEqual( value(instance.b), 2)
        self.assertEqual( value(instance.c), 2)
        self.assertEqual( value(instance.d), 2)
        # Test 2
        instance = model.create(currdir+'data14.dat', namespaces=['ns1','ns3','nsX'])
        self.assertEqual( value(instance.a), 1)
        self.assertEqual( value(instance.b), 100)
        self.assertEqual( value(instance.c), 3)
        self.assertEqual( value(instance.d), 100)
        # Test None
        instance = model.create(currdir+'data14.dat')
        self.assertEqual( value(instance.a), -1)
        self.assertEqual( value(instance.b), -2)
        self.assertEqual( value(instance.c), -3)
        self.assertEqual( value(instance.d), -4)
        #
        os.chdir(cwd)


class TestModelData(unittest.TestCase):

    def test_tableA1(self):
        """Importing a single column of data"""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'A.tab'), format='set', set='A')
        model=AbstractModel()
        model.A = Set()
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.A.data(), set(['A1', 'A2', 'A3']))

    def test_tableA2(self):
        """Importing a single column of data"""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'A.tab'))
        model=AbstractModel()
        model.A = Set()
        try:
            md.read(model)
            instance = model.create(md)
            self.fail("Should fail because no set name is specified")
        except ValueError:
            pass

    def test_tableA3(self):
        """Importing a single column of data"""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'A.tab'), set='A')
        model=AbstractModel()
        model.A = Set()
        md.read(model)
        instance = model.create(md)

    def test_tableB(self):
        """Same as test_tableA"""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'B.tab'), format='set', set='B')
        model=AbstractModel()
        model.B = Set()
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.B.data(), set([1, 2, 3]))

    def test_tableC(self):
        """Importing a multi-column table, where all columns are
        treated as values for a set with tuple values."""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'C.tab'), format='set', set='C')
        model=AbstractModel()
        model.C = Set(dimen=2)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.C.data(), set([('A1',1), ('A1',2), ('A1',3), ('A2',1), ('A2',2), ('A2',3), ('A3',1), ('A3',2), ('A3',3)]))

    def test_tableD(self):
        """Importing a 2D array of data as a set."""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'D.tab'), format='set_array', set='C')
        model=AbstractModel()
        model.C = Set(dimen=2)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.C.data(), set([('A1',1), ('A2',2), ('A3',3)]))

    def test_tableZ(self):
        """Importing a single parameter"""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'Z.tab'), format='param', param='Z')
        model=AbstractModel()
        model.Z = Param(default=99.0)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.Z.value, 1.1)

    def test_tableY(self):
        """Same as tableXW."""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'Y.tab'), index=['A'], param=['Y'])
        model=AbstractModel()
        model.A = Set(initialize=['A1','A2','A3','A4'])
        model.Y = Param(model.A)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.A.data(), set(['A1','A2','A3','A4']))
        self.assertEqual(instance.Y.extract_values(), {'A1':3.3,'A2':3.4,'A3':3.5})

    def test_tableXW_1(self):
        """Importing a table, but only reporting the values for the non-index
        parameter columns.  The first column is assumed to represent an
        index column."""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'XW.tab'), index=['A'], param=['X','W'])
        model=AbstractModel()
        model.A = Set(initialize=['A1','A2','A3','A4'])
        model.X = Param(model.A)
        model.W = Param(model.A)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.A.data(), set(['A1','A2','A3','A4']))
        self.assertEqual(instance.X.extract_values(), {'A1':3.3,'A2':3.4,'A3':3.5})
        self.assertEqual(instance.W.extract_values(), {'A1':4.3,'A2':4.4,'A3':4.5})

    def test_tableXW_2(self):
        """Like test_tableXW_1, except that set A is not defined."""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'XW.tab'), index=['A'], param=['X','W'])
        model=AbstractModel()
        model.A = Set(initialize=['A1','A2','A3'])
        model.X = Param(model.A)
        model.W = Param(model.A)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.X.extract_values(), {'A1':3.3,'A2':3.4,'A3':3.5})
        self.assertEqual(instance.W.extract_values(), {'A1':4.3,'A2':4.4,'A3':4.5})

    def test_tableXW_3(self):
        """Like test_tableXW_1, except that set A is defined in the import statment."""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'XW.tab'), index_name='A', index=['A'], param=['X','W'])
        model=AbstractModel()
        model.A = Set()
        model.X = Param(model.A)
        model.W = Param(model.A)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.A.data(), set(['A1','A2','A3']))
        self.assertEqual(instance.X.extract_values(), {'A1':3.3,'A2':3.4,'A3':3.5})
        self.assertEqual(instance.W.extract_values(), {'A1':4.3,'A2':4.4,'A3':4.5})

    def test_tableXW_4(self):
        """Like test_tableXW_1, except that set A is defined in the import statment and all values are mapped."""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'XW.tab'), index_name='B', index=['A'], param=['X','W'], param_name={'X':'R', 'W':'S'})
        model=AbstractModel()
        model.B = Set()
        model.R = Param(model.B)
        model.S = Param(model.B)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.B.data(), set(['A1','A2','A3']))
        self.assertEqual(instance.R.extract_values(), {'A1':3.3,'A2':3.4,'A3':3.5})
        self.assertEqual(instance.S.extract_values(), {'A1':4.3,'A2':4.4,'A3':4.5})

    def test_tableT(self):
        """Importing a 2D array of parameters that are transposed."""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'T.tab'), format='transposed_array', param='T')
        model=AbstractModel()
        model.B = Set(initialize=['I1','I2','I3','I4'])
        model.A = Set(initialize=['A1','A2','A3'])
        model.T = Param(model.A, model.B)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.T.extract_values(), {('A2', 'I1'): 2.3, ('A1', 'I2'): 1.4, ('A1', 'I3'): 1.5, ('A1', 'I4'): 1.6, ('A1', 'I1'): 1.3, ('A3', 'I4'): 3.6, ('A2', 'I4'): 2.6, ('A3', 'I1'): 3.3, ('A2', 'I3'): 2.5, ('A3', 'I2'): 3.4, ('A2', 'I2'): 2.4, ('A3', 'I3'): 3.5})

    def test_tableU(self):
        """Importing a 2D array of parameters."""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'T.tab'), format='array', param='U')
        model=AbstractModel()
        model.A = Set(initialize=['I1','I2','I3','I4'])
        model.B = Set(initialize=['A1','A2','A3'])
        model.U = Param(model.A, model.B)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.U.extract_values(), {('I2', 'A1'): 1.4, ('I3', 'A1'): 1.5, ('I3', 'A2'): 2.5, ('I4', 'A1'): 1.6, ('I3', 'A3'): 3.5, ('I1', 'A2'): 2.3, ('I4', 'A3'): 3.6, ('I1', 'A3'): 3.3, ('I4', 'A2'): 2.6, ('I2', 'A3'): 3.4, ('I1', 'A1'): 1.3, ('I2', 'A2'): 2.4})

    def test_tableS(self):
        """Importing a table, but only reporting the values for the non-index
        parameter columns.  The first column is assumed to represent an
        index column.  A missing value is represented in the column data."""
        md = ModelData()
        md.add(os.path.abspath(example_dir+'S.tab'), index=['A'], param=['S'])
        model=AbstractModel()
        model.A = Set(initialize=['A1','A2','A3','A4'])
        model.S = Param(model.A)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.A.data(), set(['A1','A2','A3','A4']))
        self.assertEqual(instance.S.extract_values(), {'A1':3.3,'A3':3.5})

    def test_tablePO(self):
        """Importing a table that has multiple indexing columns"""
        pyutilib.misc.setup_redirect(currdir+'importPO.dat')
        print("import "+os.path.abspath(example_dir+'PO.tab')+" : J=[A,B] P O;")
        pyutilib.misc.reset_redirect()
        md = ModelData()
        md.add(os.path.abspath(example_dir+'PO.tab'), index_name='J', index=['A','B'], param=['P','O'])
        model=AbstractModel()
        model.J = Set(dimen=2)
        model.P = Param(model.J)
        model.O = Param(model.J)
        md.read(model)
        instance = model.create(md)
        self.assertEqual(instance.J.data(), set([('A3', 'B3'), ('A1', 'B1'), ('A2', 'B2')]) )
        self.assertEqual(
            instance.P.extract_values(),
            {('A3', 'B3'): 4.5, ('A1', 'B1'): 4.3, ('A2', 'B2'): 4.4} )
        self.assertEqual(instance.O.extract_values(), {('A3', 'B3'): 5.5, ('A1', 'B1'): 5.3, ('A2', 'B2'): 5.4})
        os.remove(currdir+'importPO.dat')


if __name__ == "__main__":
    unittest.main()
